""" Module for Gemini/GNIRS specific codes
"""
import numpy as np

from pypeit import msgs
from pypeit import telescopes
from pypeit.core import framematch
from pypeit.core.wavecal import wvutils
from pypeit.par import pypeitpar
from pypeit.spectrographs import spectrograph
from pkg_resources import resource_filename
from IPython import embed


from pypeit import debugger

class GeminiGNIRSSpectrograph(spectrograph.Spectrograph):
    """
    Child to handle Gemini/GNIRS specific code
    """
    def __init__(self):
        # Get it started
        super(GeminiGNIRSSpectrograph, self).__init__()
        self.spectrograph = 'gemini_gnirs'
        self.telescope = telescopes.GeminiNTelescopePar()
        self.camera = 'GNIRS'
        self.dispname = None # TODO We need a model for setting setup specific parameters in spectrograph
        self.numhead = 2
        self.detector = [
                # Detector 1
                pypeitpar.DetectorPar(
                            dataext         = 1,
                            specaxis        = 0,
                            specflip=True,
                            spatflip=True,
                            xgap            = 0.,
                            ygap            = 0.,
                            ysize           = 1.,
                            platescale      = 0.15,
                            darkcurr        = 0.15,
                            saturation      = 150000.,
                            nonlinear       = 0.71,
                            numamplifiers   = 1,
                            gain            = 13.5,
                            ronoise         = 7.0,
                            datasec         = '[:,:]',#'[1:1024,1:1022]',
                            oscansec        = '[:,:]',#'[1:1024,1:1022]'
                            )]
        # Uses default timeunit
        # Uses default primary_hdrext
        # self.sky_file = ?
    @property
    def pypeline(self):
        return 'Echelle'


    def default_pypeit_par(self):
        """
        Set default parameters for Gemini GNIRS reductions.
        """
        par = pypeitpar.PypeItPar()
        par['rdx']['spectrograph'] = 'gemini_gnirs'
        # Frame numbers
        par['calibrations']['standardframe']['number'] = 1
        par['calibrations']['biasframe']['number'] = 0
        par['calibrations']['pixelflatframe']['number'] = 5
        par['calibrations']['traceframe']['number'] = 5
        par['calibrations']['arcframe']['number'] = 1
        # No overscan
        for key in par['calibrations'].keys():
            if 'frame' in key:
                par['calibrations'][key]['process']['overscan'] = 'none'

        # Flats
        par['calibrations']['flatfield']['illumflatten'] = False
        par['calibrations']['flatfield']['tweak_slits_thresh'] = 0.90
        par['calibrations']['flatfield']['tweak_slits_maxfrac'] = 0.10

        # Finding objects
        par['reduce']['skysub']['bspline_spacing'] = 0.8
        par['reduce']['findobj']['sig_thresh'] = 5.0
        par['reduce']['findobj']['find_trim_edge'] = [2,2]    # Slit is too short to trim 5,5 especially
        par['reduce']['findobj']['find_cont_fit'] = False     # Don't continuum fit objfind for narrow slits
        par['reduce']['findobj']['find_npoly_cont'] = 0       # Continnum order for determining thresholds
        # Extraction
        par['reduce']['extraction']['model_full_slit'] = True  # local sky subtraction operates on entire slit
        # Sky Subtraction
        par['reduce']['skysub']['global_sky_std']  = False # Do not perform global sky subtraction for standard stars
        par['reduce']['skysub']['no_poly'] = True         # Do not use polynomial degree of freedom for global skysub


        # Do not correct for flexure
        par['flexure'] = None
        # Set the default exposure time ranges for the frame typing
        par['calibrations']['pixelflatframe']['exprng'] = [None, 30]
        par['calibrations']['traceframe']['exprng'] = [None, 30]
        par['calibrations']['standardframe']['exprng'] = [None, 30]
        par['scienceframe']['exprng'] = [30, None]

        # Do not bias subtract
        #par['scienceframe']['useframe'] = 'overscan'
        # This is a hack for now until we can specify for each image type what to do. Bias currently
        # controls everything
        par['calibrations']['biasframe']['useframe'] = 'none'

        # Sensitivity function parameters
        par['sensfunc']['algorithm'] = 'IR'
        par['sensfunc']['polyorder'] = 8
        par['sensfunc']['IR']['telgridfile'] = resource_filename('pypeit', '/data/telluric/TelFit_MaunaKea_3100_26100_R20000.fits')

        return par

    def config_specific_par(self, scifile, inp_par=None):
        """
        Modify the PypeIt parameters to hard-wired values used for specific instrument configurations.


        Args:
           scifile (str):
                File to use when determining the configuration and how
                to adjust the input parameters.
           inp_par (:class:`pypeit.par.parset.ParSet`, optional):
                Parameter set used for the full run of PypeIt.  If None,
                use :func:`default_pypeit_par`.

        Returns:
           :class:`pypeit.par.parset.ParSet`: The PypeIt paramter set  adjusted for configuration specific parameter values.
        """

        par = self.default_pypeit_par() if inp_par is None else inp_par

        # TODO This is a hack for now until we figure out how to set dispname and other meta information in the
        # spectrograph class itself
        self.dispname = self.get_meta_value(scifile, 'dispname')
        # 32/mmSB_G5533 setup, covering XYJHK with short blue camera
        if '32/mm' in self.dispname:
            # Edges
            par['calibrations']['slitedges']['edge_thresh'] = 20.
            par['calibrations']['slitedges']['trace_thresh'] = 10.
            par['calibrations']['slitedges']['fit_order'] = 5
            par['calibrations']['slitedges']['max_shift_adj'] = 0.5
            par['calibrations']['slitedges']['fit_min_spec_length'] = 0.5
            par['calibrations']['slitedges']['left_right_pca'] = True
            par['calibrations']['slitedges']['pca_order'] = 3

            # Wavelengths
            par['calibrations']['wavelengths']['rms_threshold'] = 1.0  # Might be grating dependent..
            par['calibrations']['wavelengths']['sigdetect'] = 5.0
            par['calibrations']['wavelengths']['lamps'] = ['OH_GNIRS']
            par['calibrations']['wavelengths']['nonlinear_counts'] = self.detector[0]['nonlinear'] * self.detector[0][
                'saturation']
            par['calibrations']['wavelengths']['n_first'] = 2
            par['calibrations']['wavelengths']['n_final'] = [1, 3, 3, 3, 3, 3]

            # Reidentification parameters
            par['calibrations']['wavelengths']['method'] = 'reidentify'
            par['calibrations']['wavelengths']['cc_thresh'] = 0.6
            par['calibrations']['wavelengths']['reid_arxiv'] = 'gemini_gnirs.fits'
            par['calibrations']['wavelengths']['ech_fix_format'] = True
            # Echelle parameters
            # JFH This is provisional these IDs should be checked.
            par['calibrations']['wavelengths']['echelle'] = True
            par['calibrations']['wavelengths']['ech_nspec_coeff'] = 3
            par['calibrations']['wavelengths']['ech_norder_coeff'] = 5
            par['calibrations']['wavelengths']['ech_sigrej'] = 3.0

            # Tilts
            par['calibrations']['tilts']['tracethresh'] = [5.0, 10, 10, 10, 10, 10]
            par['calibrations']['tilts']['sig_neigh'] = 5.0
            par['calibrations']['tilts']['nfwhm_neigh'] = 2.0
        # 10/mmLBSX_G5532 setup, covering YJHK with the long blue camera and SXD prism
        elif '10/mmLBSX' in self.dispname:
            # Edges
            par['calibrations']['slitedges']['edge_thresh'] = 20.
            par['calibrations']['slitedges']['trace_thresh'] = 10.
            par['calibrations']['slitedges']['fit_order'] = 2
            par['calibrations']['slitedges']['max_shift_adj'] = 0.5
            par['calibrations']['slitedges']['det_min_spec_length'] = 0.20
            par['calibrations']['slitedges']['fit_min_spec_length'] = 0.20
            par['calibrations']['slitedges']['left_right_pca'] = True # Actually we need a parameter to disable PCA entirely
            par['calibrations']['slitedges']['sync_predict'] = 'nearest'

            # Wavelengths
            par['calibrations']['wavelengths']['rms_threshold'] = 1.0  # Might be grating dependent..
            par['calibrations']['wavelengths']['sigdetect'] = 5.0
            par['calibrations']['wavelengths']['lamps'] = ['Ar_IR_GNIRS']
            par['calibrations']['wavelengths']['nonlinear_counts'] = self.detector[0]['nonlinear'] * self.detector[0]['saturation']
            par['calibrations']['wavelengths']['n_first'] = 2
            par['calibrations']['wavelengths']['n_final'] = [3, 3, 3, 3]
            # Reidentification parameters
            par['calibrations']['wavelengths']['method'] = 'reidentify'
            par['calibrations']['wavelengths']['cc_thresh'] = 0.6
            par['calibrations']['wavelengths']['reid_arxiv'] = 'gemini_gnirs_10mm_LBSX.fits'
            par['calibrations']['wavelengths']['ech_fix_format'] = True
            # Echelle parameters
            par['calibrations']['wavelengths']['echelle'] = True
            par['calibrations']['wavelengths']['ech_nspec_coeff'] = 3
            par['calibrations']['wavelengths']['ech_norder_coeff'] = 3
            par['calibrations']['wavelengths']['ech_sigrej'] = 3.0

            # Tilts
            par['calibrations']['tilts']['tracethresh'] = [10, 10, 10, 10]
            par['calibrations']['tilts']['sig_neigh'] = 5.0
            par['calibrations']['tilts']['nfwhm_neigh'] = 2.0
        else:
            msgs.erro('Unrecognized GNIRS dispname')

        return par

    def init_meta(self):
        """
        Generate the meta data dict
        Note that the children can add to this

        Returns:
            self.meta: dict (generated in place)

        """
        meta = {}
        # Required (core)
        meta['ra'] = dict(ext=0, card='RA')
        meta['dec'] = dict(ext=0, card='DEC')
        meta['target'] = dict(ext=0, card='OBJECT')
        meta['decker'] = dict(ext=0, card='DECKER')

        meta['binning'] = dict(ext=0, card=None, default='1,1')
        meta['mjd'] = dict(ext=0, card='MJD_OBS')
        meta['exptime'] = dict(ext=0, card='EXPTIME')
        meta['airmass'] = dict(ext=0, card='AIRMASS')
        # Extras for config and frametyping
        meta['dispname'] = dict(ext=0, card='GRATING')
        meta['hatch'] = dict(ext=0, card='COVER')
        meta['dispangle'] = dict(ext=0, card='GRATTILT', rtol=1e-4)
        meta['idname'] = dict(ext=0, card='OBSTYPE')

        # Ingest
        self.meta = meta

    def configuration_keys(self):
        return ['decker', 'dispname', 'dispangle']

    def check_frame_type(self, ftype, fitstbl, exprng=None):
        """
        Check for frames of the provided type.
        """
        good_exp = framematch.check_frame_exptime(fitstbl['exptime'], exprng)
        if ftype == 'science':
            return good_exp & (fitstbl['idname'] == 'OBJECT')
        if ftype == 'standard':
            return good_exp & (fitstbl['idname'] == 'OBJECT')
        if ftype in ['pixelflat', 'trace']:
            # Flats and trace frames are typed together
            return good_exp & (fitstbl['idname'] == 'FLAT')
        if ftype in ['pinhole', 'dark', 'bias']:
            # Don't type pinhole, dark, or bias frames
            return np.zeros(len(fitstbl), dtype=bool)
        if ftype in ['arc', 'tilt']:
            return good_exp & (fitstbl['idname'] == 'ARC')

        msgs.warn('Cannot determine if frames are of type {0}.'.format(ftype))
        return np.zeros(len(fitstbl), dtype=bool)

#    def parse_binning(self, inp, det=1):
#        return '1,1'

    def order_platescale(self, order_vec, binning=None):


        """
        Returns the plate scale in arcseconds for each order

        Parameters
        ----------
        None

        Optional Parameters
        --------------------
        binning: str

        Returns
        -------
        order_platescale: ndarray, float

        """
        if '10/mmLBSX' in self.dispname:
            return np.full(order_vec.size, 0.05)
        elif '32/mm' in self.dispname:
            return np.full(order_vec.size, 0.15)
        else:
            msgs.error('Unrecognized disperser')

    @property
    def norders(self):
        if '10/mmLBSX' in self.dispname:
            return 4
        elif '32/mm' in self.dispname:
            return 6
        else:
            msgs.error('Unrecognized disperser')

    @property
    def order_spat_pos(self):
        if '10/mmLBSX' in self.dispname:
            return np.array([0.050, 0.215, 0.442, 0.759])
        elif '32/mm' in self.dispname:
            return np.array([0.2955097 , 0.37635756, 0.44952223, 0.51935601, 0.59489503, 0.70210309])
        else:
            msgs.error('Unrecognized disperser')

    @property
    def orders(self):
        if '10/mmLBSX' in self.dispname:
            return np.arange(6,2,-1, dtype=int)
        elif '32/mm' in self.dispname:
            return np.arange(8,2,-1,dtype=int)
        else:
            msgs.error('Unrecognized disperser')

    @property
    def spec_min_max(self):
        if '10/mmLBSX' in self.dispname:
            spec_max = np.asarray([1022, 1022, 1022, 1022])
            spec_min = np.asarray([450, 0, 0, 0])
            return np.vstack((spec_min, spec_max))
        elif '32/mm' in self.dispname:
            spec_max = np.asarray([1022, 1022, 1022, 1022, 1022, 1022])
            spec_min = np.asarray([512, 280, 0, 0, 0, 0])
            return np.vstack((spec_min, spec_max))
        else:
            msgs.error('Unrecognized disperser')



    def bpm(self, filename, det, shape=None, msbias=None):
        """
        Override parent bpm function with BPM specific to Gemini/GNIRS

        .. todo::
            Allow for binning changes.

        Parameters
        ----------
        det : int, REQUIRED
        **null_kwargs:
            Captured and never used

        Returns
        -------
        bpix : ndarray
          0 = ok; 1 = Mask

        """
        msgs.info("Custom bad pixel mask for GNIRS")
        bpm_img = self.empty_bpm(filename, det, shape=shape)

        # Fill in bad pixels if a master bias frame is provided
        if msbias is not None:
            return self.bpm_frombias(msbias, det, bpm_img)

        # JFH Changed. Dealing with detector scratch
        if det == 1:
            bpm_img[687:765,12:16] = 1.
            bpm_img[671:687,8:13] = 1.
        #    bpm_img[:, 1000:] = 1.

        return bpm_img






"""
Module to generate an optical model for a spectrograph.
"""
import warnings
import numpy

# ----------------------------------------------------------------------
# General class for a reflection grating
class ReflectionGrating:
    """
    Doc this!
    """
    def __init__(self, ruling, tilt, roll, yaw, central_wave=None):
        self.ruling = ruling            # Grating groove spacing in lines per mm
        self.tilt = tilt                # Effective grating tilt (deg)
        self.roll = roll                # Grating roll (deg)
        self.yaw = yaw                  # Grating yaw (deg)

        self.central_wave = central_wave    # Central wavelength (angstroms)

        self.transform = self._get_grating_transform()

    def _get_grating_transform(self):
        """
        Taken from xidl/DEEP2/spec2d/pro/model/setup.pro ; same as in
        xidl/DEEP2/spec2d/pro/model/gsetup.pro

        Here tilt is the same as mu (!MU), roll is the same as roll3
        (!GR_YERR), and yaw is the same as o3 (!GR_ZERR).
        
        Assumes phin=0. (ie adopts the roll/yaw approach)

        MIRROR/GRATING: Better in thetax, thetay description (see above)
        for GRATING: need to add the third rotation Note the hack with
        phin + PI.  This is needed to keep the transformed x, y axes
        from flipping, wrt the original x,y.  Not a problem wrt
        reflections but if we want to work _within_ a system it is a
        problem. Cf. camera also note that this _forces_ us to work in a
        particular hemisphere, and thus we must make use of the negative
        theta as needed.
        """
        theta = -numpy.radians(self.tilt)
        rho = numpy.radians(self.roll)
        xsi = numpy.radians(self.yaw)

        cost = numpy.cos(theta)
        sint = numpy.sin(theta)
        cosr = numpy.cos(rho)
        sinr = numpy.sin(rho)
        cosx = numpy.cos(xsi)
        sinx = numpy.sin(xsi)

        return numpy.array([[ cosx*cosr,  sint*sinr+cost*sinx*cosr, -cost*sinr+sint*sinx*cosr],
                            [     -sinx,                 cost*cosx,                 sint*cosx],
                            [ cosx*sinr, -sint*cosr+cost*sinx*sinr,  cost*cosr+sint*sinx*sinr]])

    def reflect(self, r, wave=None, order=1):
        """
        Propagate an input ray for a given wavelength and order.

        wave is in angstroms
        ruling is in mm^-1

        If more than one wave provided, wavelength samples are ordered
        along the first axis.

        Taken from xidl/DEEP2/spec2d/pro/model/qmodel.pro.
        """
        if wave is None and self.central_wave is None:
            raise ValueError('Must define a wavelength for the calculation.')
        if wave is None:
            warnings.warn('Using central wavelength for calculation.')
        _wave = numpy.array([self.central_wave]) if wave is None else numpy.atleast_1d(wave)
        if _wave.ndim > 1:
            raise NotImplementedError('Input wavelength must be one number or a vector.')
        nwave = _wave.size

        _r = numpy.atleast_2d(r)
        if _r.ndim > 2:
            raise NotImplementedError('Rays must be 1D for a single ray, or 2D for multiple.')

        # Transform into the grating conjugate surface
        _r = OpticalModel.conjugate_surface_transform(_r, self.transform, forward=True)

        # Get the grating input angles
        alpha = -numpy.arctan2(-_r[:,1],-_r[:,2])
        gamma = numpy.arctan2(_r[:,0],numpy.sqrt(numpy.square(_r[:,1])+numpy.square(_r[:,2])))

        # Use the grating equation to get the output angle (minus sign
        # in front of sin(alpha) is for reflection)
        beta = numpy.arcsin((order * 1e-7 * self.ruling * _wave[None,:] / numpy.cos(gamma)[:,None])
                            - numpy.sin(alpha)[:,None])

        # Revert to ray vectors
        wavesign = 1-2*(_wave < 0)
        cosg = numpy.cos(gamma)
        _r = numpy.array([numpy.repeat(numpy.sin(gamma), nwave).reshape(-1,nwave),
                          numpy.sin(-beta*wavesign[None,:])*cosg[:,None],
                          numpy.cos(-beta*wavesign[None,:])*cosg[:,None] ]).T

        if nwave == 1:
            # Flatten if only one wave provided
            _r = _r[0]

        # Return vectors transformed out of the grating conjugate surface
        return OpticalModel.conjugate_surface_transform(_r, self.transform)


# ----------------------------------------------------------------------
# Vanilla imaging spectrograph optical model
class OpticalModel:
    """
    Vanilla optical model for an imaging spectrograph.
    
    Model includes four elements:
        - Slit mask at the focal plane
        - Reflective Collimator
        - Reflective Grating
        - Refractive Camera

    Primary objective is to trace light rays from the focal plane
    position to a position in the imagine coordinate system of the
    camera.  See :func:`mask_to_imaging_coordinates`.

    It is expected that each spectrograph will have its own optical
    model to perturb what is done in the vanilla model as necessary.

    .. todo:
        - Provide a ParSet for the arguments of the function?
        - I say this is vanilla, but I'm not sure how much of this is
          still DEIMOS specific.

    Args:
        pupil_distance (float):
            Pupil distance in mm
        focal_r_surface (float):
            Radius of the image surface in mm
        focal_r_curvature (float):
            Focal-plane radius of curvature in mm
        mask_r_curvature (float):
            Mask radius of curvature in mm
        mask_tilt_angle (float):
            Mask tilt angle in radians
        mask_y_zeropoint (float):
            Mask y zero point in mm
        mask_z_zeropoint (float):
            Mask z zero point in mm
        collimator_d (float):
            Collimator distance in mm
        collimator_r (float):
            Collimator radius of curvature in mm
        collimator_k (float):
            Collimator curvature constant
        coll_tilt_err (float):
            Collimator tilt error in radians
        coll_tilt_phi (float):
            Collimator tilt phi in radians
        grating (:class:`ReflectionGrating`):
            Grating object that evaluates the grating equation and
            performs the grating reflection.
        camera_tilt (float):
            Camera angle in radians
        camera_phi (float):
            Camera tilt phi angle in radians
        camera_focal_length (float):
            Camera focal length in mm
        camera_distortions (object):
            Class to apply/remove camera distortions.  Can be None.  If
            provided, must have `remove_distortion` and
            `apply_distortion` methods.
        imaging_rotation (float):
            Image coordinate system rotation in radians
        optical_axis (numpy.ndarray):
            Camera optical axis center (x,y) in mm

    Attributes:
    """
    def __init__(self, pupil_distance, focal_r_surface, focal_r_curvature, mask_r_curvature,
                 mask_tilt_angle, mask_y_zeropoint, mask_z_zeropoint, collimator_d, collimator_r,
                 collimator_k, coll_tilt_err, coll_tilt_phi, grating, camera_tilt, camera_phi,
                 camera_focal_length, camera_distortions, imaging_rotation, optical_axis):

        # TODO: Are focal_r_surface and focal_r_curvature supposed to be
        # the same?
        self.pupil_distance = pupil_distance        # Pupil distance in mm
        self.focal_r_surface = focal_r_surface      # Radius of the image surface in mm
        self.focal_r_curvature = focal_r_curvature  # Focal-plane radius of curvature in mm

        self.mask_r_curvature = mask_r_curvature    # Mask radius of curvature in mm
        self.mask_tilt_angle = mask_tilt_angle      # Mask tilt angle in radians
        self.mask_y_zeropoint = mask_y_zeropoint    # Mask y zero point in mm
        self.mask_z_zeropoint = mask_z_zeropoint    # Mask z zero-point in mm

        self.collimator_d = collimator_d            # Collimator distance in mm
        self.collimator_r = collimator_r            # Collimator radius of curvature in mm
        self.collimator_k = collimator_k            # Collimator curvature constant
        self.coll_tilt_err = coll_tilt_err          # Collimator tilt error in radians
        self.coll_tilt_phi = coll_tilt_phi          # Collimator tilt phi in radians

        self.collimator_transform = self._collimator_transform()

        self.grating = grating                      # Grating, must be a Grating object

        self.camera_tilt = camera_tilt              # Camera angle in radians
        self.camera_phi = camera_phi                # Camera tilt phi angle in radians
        self.camera_focal_length = camera_focal_length  # Camera focal length in mm

        # Camera transmission transform is the same as a refelction
        # transform with the opposite angle: tranformation from
        # theta-x,y is better, although this is not ridiculous.
        self.camera_transform \
                = OpticalModel.get_reflection_transform(-self.camera_tilt,
                                                        self.camera_phi + numpy.pi)

        self.camera_distortions = camera_distortions    # Class to apply/remove camera distortions

        self.imaging_rotation = imaging_rotation    # Image coordinate system rotation in radians
        self.optical_axis = optical_axis            # Camera optical axis center (x,y) in mm

    def _collimator_transform(self):
        """
        tilt and tilt_error are in radians

        Taken from xidl/DEEP2/spec2d/pro/model/setup.pro:

        COLL ERROR (first order): In this case, we must remove the phi
        we put in, hence the more complex form
        """
        theta = 2 * self.coll_tilt_err
        phi = self.coll_tilt_phi + numpy.pi/2

        cost = numpy.cos(theta)
        sint = numpy.sin(theta)
        cosp = numpy.cos(phi)
        sinp = numpy.sin(phi)

        # TODO: Is this right?  It's what's in the IDL version...
        c = 1-(1-cost)*cosp*cosp * sinp*sinp + cost*cosp*cosp

        return numpy.array([[ 1-(1-cost)*sinp*sinp,   (1-cost)*cosp*sinp, -sinp*sint ],
                            [   (1-cost)*cosp*sinp,                    c,  cosp*sint ],
                            [            sint*sinp,           -sint*cosp,       cost ]])

    @staticmethod
    def conjugate_surface_transform(ray, surface_transform, forward=False):
        """
        Transform a ray by a surface.

        Taken from xidl/DEEP2/spec2d/pro/model/gen_xfm.pro

        .. todo::
            I'm trying to mimic what was done in the IDL code,
            meaning the matrix ordering may not make sense...

        Args:
            ray (numpy.ndarray):
                The rays to tranform.  If more than one ray is provided,
                they must be organized along the last axis.  I.e., for
                two rays, the shape of ray should be (2,3).
            surface_transform (numpy.ndarray):
                The transform for the surface.  For example, see
                :func:`OpticalModel.get_reflection_transform`.  For more
                than one surface, the surface matrix must be organized
                along the last two axes.  I.e., for two surfaces, the
                shape of surface should be (2,3,3).
        Returns:
            numpy.ndarray: The array with the reflected arrays.
        """
        return numpy.einsum('...ji,...i', surface_transform, ray) if forward else \
                    numpy.einsum('...ij,...i', surface_transform, ray)

    @staticmethod
    def reflect(r, surface):
        """
        Reflect a (set of) ray(s) of a surface using the provided
        transformation.

        .. todo::
            (KBW) I'm trying to mimic what was done in the IDL code,
            meaning the matrix ordering may not make sense...

        Args:
            r (numpy.ndarray):
                The rays to reflect of the defined surface.  If more
                than one ray is provided, they must be organized along
                the last axis.  I.e., for two rays, the shape of r
                should be (2,3).
            surface (numpy.ndarray):
                The transform for the surface used in the reflection.
                For example, see
                :func:`OpticalModel.get_reflection_transform`.  For more
                than one surface, the surface matrix must be organized
                along the last two axes.  I.e., for two surfaces, the
                shape of surface should be (2,3,3).

        Returns:
            numpy.ndarray: The array with the reflected arrays.
        """
        _r = OpticalModel.conjugate_surface_transform(r, surface, forward=True)
        _r[...,2] *= -1
        return OpticalModel.conjugate_surface_transform(_r, surface)

    @staticmethod
    def get_reflection_transform(theta, phi):
        """
        General reflection transform.

        Taken from xidl/DEEP2/spec2d/pro/model/setup.pro:
        """
        cosp = numpy.cos(phi+numpy.pi/2)
        sinp = numpy.sin(phi+numpy.pi/2)
        cost = numpy.cos(theta)
        sint = numpy.sin(theta)
        transform = numpy.array([[      cosp,       sinp, numpy.zeros_like(phi)],
                                 [-cost*sinp,  cost*cosp,                  sint],
                                 [ sint*sinp, -sint*cosp,                  cost]])

        # If theta and phi are vectors move the axes so that, e.g.,
        # transform[0] is the transform for (theta[0],phi[0])
        return numpy.moveaxis(transform, -1, 0) if transform.ndim > 2 else transform

    # TODO: Inverse operation not provided; see xidl/DEEP2/spec2d/pro/model/proj_to_mask.pro
    def project_mask_onto_plane(self, x, y, a):
        """
        Project slitmask coords (curved surface) onto a plane.
    
        Generic as long as self is properly defined
    
        Taken from xidl/DEEP2/spec2d/pro/model/mask_to_proj.pro:
    
        This is pure geometry.
    
        Args:
            x (float):
                x coordinate on mask surface in mm
            y (float):
                y coordinate on mask surface in mm
            a (float):
                position angle on curved mask surface
    
        Returns:
            float: Three floats, the x, y, and position angle of the
            coordinate in the (focal-)planar system
        """
        cosm = numpy.cos(x / self.mask_r_curvature)
        sint = numpy.sin(self.mask_tilt_angle)
        cost = numpy.cos(self.mask_tilt_angle)
        tant = sint/cost

        px = self.mask_r_curvature * numpy.sin(x / self.mask_r_curvature)
        py = (y - self.mask_r_curvature * tant * (1. - cosm)) * cost + self.mask_y_zeropoint
        pa = numpy.arctan((numpy.tan(numpy.radians(a)) - tant * px / self.mask_r_curvature)
                            * cost / cosm)
    
        # What follows is a small correction for the fact that the mask does
        # not lie exactly in the spherical image surface (where the
        # distortion values and gnomonic projection are calculated) and the
        # rays are arriving from the pupil image; thus, the exact locations
        # are moved _slightly_ wrt the telescope optical axis.  Note also
        # that these corrections are only calculated to first order.
        # TODO: Make this an optional correction?
        rho = numpy.sqrt(numpy.square(px) + numpy.square(py))
    
        # Spherical image and mask surface heights
        hs = self.focal_r_surface * (1. - numpy.sqrt(1. - numpy.square(rho/self.focal_r_surface)))
        hm = self.mask_z_zeropoint + y * sint + self.mask_r_curvature*(1. - cosm)
    
        px *= (1 - (hs - hm) / self.pupil_distance)
        py *= (1 - (hs - hm) / self.pupil_distance)

        return px, py, pa
    
    def telescope_plane_coo_to_unit_vector(self, x, y): #, a):
        """
        Convert the coordinates in the focal plane to the ray tracing unit
        vector.
    
        !! IGNORES ANGLE !!
    
        Taken from xidl/DEEP2/spec2d/pro/model/pre_grating.pro
        """
        r2 = numpy.square(x) + numpy.square(y)
        hm = self.focal_r_curvature - numpy.sqrt(numpy.square(self.focal_r_curvature)-r2)
        theta = numpy.arctan(numpy.sqrt(r2) / (self.pupil_distance - hm))
        phi = numpy.arctan2(y, x)
    
        sint = numpy.sin(theta)
        return numpy.array([ numpy.cos(phi)*sint, numpy.sin(phi)*sint, numpy.cos(theta) ]).T

    def telescope_plane_coo_to_collimator_angle(self, x, y):
        """
        Convert the coordinates in the focal plane to the collimator
        angles.

        Taken from xidl/DEEP2/spec2d/pro/model/coll_angle.pro
        """
        r2 = numpy.square(x) + numpy.square(y)
        fr2 = numpy.square(self.focal_r_curvature)
        hm = self.focal_r_curvature - numpy.sqrt(fr2-r2)
        d2 = self.pupil_distance + self.collimator_d

        cott = (self.pupil_distance - hm) / numpy.sqrt(r2)
        cott2 = numpy.square(cott)
        k = 1. + (1. + self.collimator_k) * cott2
        d = d2 * (1. + self.collimator_k)

        # The following is general for conic sections.  In practice, a
        # parabola is fine.  Note the switch in sign for quadratic root
        b = numpy.sqrt(cott2 + d2*k*(2.*self.collimator_r - d) \
                    / numpy.square(self.collimator_r - d))
        r = (self.collimator_r - d) * (b - cott) / k if self.collimator_r > d \
                    else (d - self.collimator_r) * (b + cott) / k

        # This is for a parabola:
        # r = !R_COLL * (sqrt (cott*cott + 2. * d2 / !R_COLL ) - cott)
    
        # The general conic form (important)
        r /= self.collimator_r

        return numpy.arctan(r / numpy.sqrt(1. - (1.+self.collimator_k)*r*r)), numpy.arctan2(y, x)

    def mask_coo_to_grating_input_vectors(self, x, y):
        """
        Propagate rays from the mask plane to the grating.

        Taken from xidl/DEEP2/spec2d/pro/model/pre_grating.pro
        """
        _x = numpy.atleast_1d(x)
        _y = numpy.atleast_1d(y)

        # Project the mask coordinates onto the telescope focal plane
        px, py, pa = self.project_mask_onto_plane(_x, _y, numpy.zeros_like(_x))

        # Project the focal-plane coordinates onto the collimator
        theta, phi = self.telescope_plane_coo_to_collimator_angle(px, py)

        # Form the reflection transform (a1)
        collimator_reflection = OpticalModel.get_reflection_transform(theta, phi)

        # Get the telescope plane unit vectors
        r = self.telescope_plane_coo_to_unit_vector(px, py)

        # Reflect off the collimator
        r = OpticalModel.reflect(r, collimator_reflection)

        # Transform with collimator error (e1)
        return OpticalModel.conjugate_surface_transform(r, self.collimator_transform,
                                                        forward=True)

    def ics_coo_to_grating_output_vectors(self, x, y):
        """
        Revert rays from the CCD coordinates back to the grating
        output vectors.

        INPUT IS MM

        Taken from xidl/DEEP2/spec2d/pro/model/ics_post_grating.pro
        """
        # 8. Remove the mosaic transform
        cosp = numpy.cos(-self.imaging_rotation)
        sinp = numpy.sin(-self.imaging_rotation)
        _x = x - self.optical_axis[0]
        _y = y - self.optical_axis[1]
        xp =  cosp * _x + sinp * _y
        yp = -sinp * _x + cosp * _y

        # 7. Display coords are different than the x,y,z convention adopted so far:
        xp *= -1
        yp *= -1

        # 6. Convert to x,y in focal plane:
        rp = numpy.sqrt(xp*xp + yp*yp)
        theta = numpy.arctan(rp / self.camera_focal_length)
        phi = numpy.arctan2(yp, xp)

        # 5. Remove the camera distortion in theta (phi unchanged)
        if self.camera_distortions is not None:
            theta = self.camera_distortions.remove_distortion(theta)

        # 4. Convert angles into unit vectors
        sint = numpy.sin(theta)
        r = numpy.array([numpy.cos(phi)*sint, numpy.sin(phi)*sint, numpy.cos(theta)]).T

        # 3. Transform through the camera system
        return OpticalModel.conjugate_surface_transform(r, self.camera_transform)

    def grating_output_vectors_to_ics_coo(self, r, sign=1):
        """
        Revert rays from the CCD coordinates back to the grating
        output vectors.

        There's a sign degeneracy going this way, so it must be defined.

        OUTPUT IS MM

        Inverted xidl/DEEP2/spec2d/pro/model/ics_post_grating.pro
        """
        # 3. Transform through the camera system.
        _r = OpticalModel.conjugate_surface_transform(r, self.camera_transform, forward=True)

        # 4. Get angles from unit vectors
        theta = numpy.arccos(_r[...,2])
        phi = numpy.arctan2(_r[...,1], _r[...,0])

        # 5. Apply the camera distortion in theta (phi unchanged)
        if self.camera_distortions is not None:
            theta = self.camera_distortions.apply_distortion(theta)

        # 6. Convert angles to x,y
        # WARNING: There's a +/- 90 deg limit on imaging_rotation when
        # this will work...
        tanp = numpy.tan(phi)
        xp = self.camera_focal_length * numpy.tan(theta) / numpy.sqrt(1 + numpy.square(tanp))
        yp = xp * tanp

        # 7. Image coordinates are flipped.
        xp *= sign
        yp *= sign

        # 8. Add the mosaic transform and return the coordinates
        cosp = numpy.cos(-self.imaging_rotation)
        sinp = numpy.sin(-self.imaging_rotation)
        return xp*cosp - yp*sinp + self.optical_axis[0], xp*sinp + yp*cosp + self.optical_axis[1]

    def mask_to_imaging_coordinates(self, x, y, wave, order):
        """
        Convert mask coordinates in mm to detector coordinates in pixels.

        wave is in angstroms

        If more than one wavelength is provided, wavelength samples are
        ordered along the first axis.

        Taken from xidl/DEEP2/spec2d/pro/model/qmodel.pro.
        """
        # First get the grating input vectors
        r = self.mask_coo_to_grating_input_vectors(x, y)

        # Reflect the rays off the grating
        r = self.grating.reflect(r, wave, order)

        # Propagate the rays through the camera to the detector and
        # return the imaging coordinates (in mm)
        return self.grating_output_vectors_to_ics_coo(r, sign=1-2*(x<0))


# ----------------------------------------------------------------------
# Detector coordinates
class DetectorMap:
    """
    General class for mapping the image plane to the pixel coordinates
    for multiple detectors in a mosaic.

    !! PIXEL COORDINATES ARE 1-INDEXED !!

    All CCDs in the detector are assumed to have the same size.

    .. todo:
        - Allow for instantiation arguments.
        - Remove ccd_gap and ccd_edge from attributes?

    Attributes:
        nccd (int):
            Number of CCD chips in the detector.
        npix (numpy.ndarray):
            Number of pixels in each dimension (x,y) of all CCDs.
        pixel_size (float):
            Pixel size in mm.
        ccd_gap (numpy.ndarray):
            The nominal gap between each chip in (x,y) in mm.
        ccd_edge (numpy.ndarray):
            The width of the CCD edge in (x,y) in mm.
        ccd_size (numpy.ndarray):
            The nominal size in number of pixels of each CCD accounting
            for gap, edge width, and number of pixels.
        ccd_center (numpy.ndarray):
            The (x,y) center of each CCD in number of pixels, accounting
            for the per CCD offset.
        rotation (numpy.ndarray):
            The rotation of each CCD.
        rot_matrix (numpy.ndarray):
            The rotation matrix used for each CCD.
    """
    def __init__(self):
        # Basic detector with vanilla properties

        # Number of chips
        self.nccd = 1

        # Number of pixels for each chip in each dimension
        # TODO: Currently assumes all chips are the same
        self.npix = numpy.array([2048, 2048])

        # The size of the CCD pixels in mm
        # TODO: Currently assumes all chips are the same
        self.pixel_size = 0.015

        # Nominal gap between each CCD in each dimension in mm
        # TODO: Currently assumes all chips are the same
        self.ccd_gap = numpy.array([0, 0])

        # Width of the CCD edge in each dimension in mm
        # TODO: Currently assumes all chips are the same
        self.ccd_edge = numpy.array([1, 1])

        # Effective size of each chip in each dimension in pixels
        # TODO: Currently assumes all chips are the same
        self.ccd_size = self.npix + (2*self.ccd_edge + self.ccd_gap)/self.pixel_size

        # Center coordinates
        origin = numpy.array([[0.,0.]])
        offset = numpy.array([[0.,0.]])
        self.ccd_center = origin * self.ccd_size[None,:] + offset
        
        # Construct the rotation matrix
        self.rotation = numpy.radians([0])
        cosa = numpy.cos(self.rotation)
        sina = numpy.sin(self.rotation)
        self.rot_matrix = numpy.array([cosa, -sina, sina, cosa]).T.reshape(self.nccd,2,2)

    def image_coordinates(self, x_pix, y_pix, detector=1, in_mm=True):
        """
        Convert the provided (1-indexed) pixel coordinates into
        coordinates in the image plane.

        Args:
            x_pix (:obj:`float` or array-like):
                Pixel coordinate in x (1-indexed) on the detector
            y_pix (:obj:`float` or array-like):
                Pixel coordinate in y (1-indexed) on the detector
            detector (:obj:`int` or array-like, optional):
                Relevant detector for the pixel coordinates.  Default is
                1.  Can be a single detector used for all coordinates or
                an array that provides the detector for each (x,y)
                coordinate.
            in_mm (:obj:`bool`, optional):
                Return the coordinates in mm.

        Returns:
            float, numpy.ndarray: Returns two objects with the x and y
            coordinates.  Return object type is based on the input.

        Raises:
            ValueError:
                Raised if the detector number is not valid or if the
                input x and y arrays do not have the same shape.
        """
        # Reshape into vectors
        _x = numpy.atleast_1d(x_pix)
        _y = numpy.atleast_1d(y_pix)
        if _x.shape != _y.shape:
            raise ValueError('Mismatch error between x and y shape.')
        
        # Allow the detector to be coordinate specific or one value for
        # all coordinates
        _d = numpy.atleast_1d(detector)-1
        if _d.shape != _x.shape and len(_d) == 1:
            _d = numpy.full(_x.shape, detector-1, dtype=int)
        if numpy.any((_d >= self.nccd) | (_d < 0)):
            raise ValueError('Incorrect detector number')

        # Keep the input shape for the returned arrays
        inp_shape = None
        if len(_x.shape) > 0:
            inp_shape = _x.shape
            _d = _d.ravel()
            _x = _x.ravel()
            _y = _y.ravel()

        # Offset by the chip center
        coo = numpy.array([_x, _y]).T - self.npix[None,:]/2

        # Rotatate and offset by the CCD center
        coo = numpy.array([numpy.matmul(self.rot_matrix[d], c) for d,c in zip(_d,coo)]) \
                    + self.ccd_center[_d,:]

        x_img = coo[0,0] if inp_shape is None else coo[:,0].reshape(inp_shape)
        y_img = coo[0,1] if inp_shape is None else coo[:,1].reshape(inp_shape)

        # Return with the appropriate shape
        return x_img * self.pixel_size if in_mm else x_img, \
                    y_img * self.pixel_size if in_mm else y_img

    def ccd_coordinates(self, x_img, y_img, in_mm=True):
        """
        Convert the provided coordinates in the image plane to
        (1-indexed) pixel coordinates on the relevant detector.

        Args:
            x_img (:obj:`float` or array-like):
                Image coordinates in x
            y_img (:obj:`float` or array-like):
                Image coordinates in y
            in_mm (:obj:`bool`, optional):
                The input coordinates are provided in mm, not pixels.

        Returns:
            float, numpy.ndarray: Returns three objects, the detector
            associated with each coordinates and the x and y pixel
            coordinates on that detector.  Return object type (float or
            array) is based on the input.

        Raises:
            ValueError:
                Raised if the input x and y arrays do not have the same
                shape.
        """

        # Reshape into vectors and convert to pixels, if necessary
        _x = numpy.atleast_1d(x_img/self.pixel_size if in_mm else x_img)
        _y = numpy.atleast_1d(y_img/self.pixel_size if in_mm else y_img)
        if _x.shape != _y.shape:
            raise ValueError('Mismatch error between x and y shape.')

        # Keep the input shape for the returned arrays
        inp_shape = None
        if len(_x.shape) > 0:
            inp_shape = _x.shape
            _x = _x.ravel()
            _y = _y.ravel()

        # Offset by the CCD center for each chip
        coo = numpy.array([_x, _y]).T[None,:,:] - self.ccd_center[:,None,:]

        # Apply the rotation matrix and offset by the chip center
        coo = numpy.array([[numpy.matmul(r.T, _c) for _c in c] \
                                for r,c in zip(self.rot_matrix, coo)]) + self.npix[None,None,:]/2

        # Determine the associated detector (1-indexed)
        indx = numpy.all((coo > 0) & (coo <= self.npix[None,None,:]), axis=2)
        on_ndet = numpy.sum(indx, axis=0)
        if numpy.any(on_ndet == 0):
            warnings.warn('Points may not be on any detector!')
        if numpy.any(on_ndet > 1):
            warnings.warn('Points may be on more than one detector!')
        d = numpy.amax(numpy.arange(self.nccd)[:,None]*indx, axis=0) + 1
        d[numpy.sum(indx, axis=0) == 0] = -1

        # Pull out the coordinates for the correct detector
        coo = numpy.array([coo[_d-1,i,:] if _d > 0 else numpy.full(coo.shape[2:],-1.) 
                                for i,_d in enumerate(d)])

        # Return the coordinates
        return d if inp_shape is None else d.reshape(inp_shape), \
                    coo[0,0] if inp_shape is None else coo[:,0].reshape(inp_shape), \
                    coo[0,1] if inp_shape is None else coo[:,1].reshape(inp_shape)





""" Object to hold + process a single image.
This module also includes the build_from_list() method
which is how the ScienceImage is most frequently generated. """

import inspect

import os
import numpy as np


from pypeit import msgs

from pypeit.core import procimg
from pypeit.par import pypeitpar
from pypeit import utils

from pypeit.images import pypeitimage
from pypeit.images import combineimage

from IPython import embed


class ScienceImage(pypeitimage.PypeItImage):
    """
    Class to generate and hold a science image

    Child of PypeItImage

    Args:
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            Spectrograph used to take the data.
        det (:obj:`int`):
            The 1-indexed detector number to process.
        par (:class:`pypeit.par.pypeitpar.ProcessImagesPar`):
            Parameters that dictate the processing of the images.  See
            :class:`pypeit.par.pypeitpar.ProcessImagesPar` for the
            defaults.
        image (np.ndarray):
        ivar (np.ndarray):
        bpm (np.ndarray):
            Bad pixel mask.  Held in ImageMask
        rn2img (np.ndarray, optional):
        crmask (np.ndarray, optional):
        mask (np.ndarray, optional):
        files (list, optional):
            List of filenames that went into the loaded image

    """
    frametype = 'science'

    def __init__(self, spectrograph, det, par, image, ivar, bpm, rn2img=None,
                 crmask=None, mask=None, files=[]):

        # Init me
        pypeitimage.PypeItImage.__init__(self, image, ivar=ivar, rn2img=rn2img,
                                         bpm=bpm, crmask=crmask, mask=mask)

        # Required attribs
        self.spectrograph = spectrograph
        if not isinstance(par, pypeitpar.ProcessImagesPar):
            msgs.error('Provided ParSet for must be type ProcessImagesPar.')
        self.par = par
        self.det = det

        # Not required
        self.files = files

    def build_crmask(self, subtract_img=None):
        """
        Call to ImageMask.build_crmask which will
        generate the cosmic ray mask

        Args:
            subtract_img (np.ndarray, optional):
                Image to be subtracted off of self.image prior to CR evaluation

        Returns:
            np.ndarray: Boolean array of self.crmask

        """
        return super(ScienceImage, self).build_crmask(self.spectrograph, self.det,
                                                      self.par, self.image,
                                                      utils.inverse(self.ivar),
                                                      subtract_img=subtract_img).copy()

    def build_mask(self, saturation=1e10, mincounts=-1e10, slitmask=None):
        """
        Call to ImageMask.build_mask()
        This generates the full Image mask

        Args:
            saturation (float, optional):
            mincounts (float, optional):
            slitmask (np.ndarray, optional):

        Returns:
            np.ndarray:  The full mask, held in self.mask

        """
        super(ScienceImage, self).build_mask(self.image, self.ivar,
                                             saturation=saturation,
                                             mincounts=mincounts,
                                             slitmask=slitmask)
        return self.mask.copy()

    def update_mask_cr(self, subtract_img=None):
        """
        Updates the CR mask values in self.mask
        through a call to ImageMask.build_crmask which
        generates a new CR mask and then a call to
        ImageMask.update_mask_cr() which updates self.mask

        Args:
            subtract_img (np.ndarray, optional):
                If provided, this is subtracted from self.image prior to
                CR masking
        """
        # Generate the CR mask (and save in self.crmask)
        super(ScienceImage, self).build_crmask(self.spectrograph, self.det,
                                               self.par, self.image,
                                               utils.inverse(self.ivar),
                                               subtract_img=subtract_img).copy()
        # Now update the mask
        super(ScienceImage, self).update_mask_cr(self.crmask)

    def __sub__(self, other):
        """
        Subtract a ScienceImage object from another
        Extras (e.g. ivar, masks) are included if they are present

        Args:
            other (ScienceImage):

        Returns:
            ScienceImage:

        """
        if not isinstance(other, ScienceImage):
            msgs.error("Misuse of the subtract method")
        # Images
        newimg = self.image - other.image

        # Mask time
        outmask_comb = (self.mask == 0) & (other.mask == 0)

        # Variance
        if self.ivar is not None:
            new_ivar = utils.inverse(utils.inverse(self.ivar) + utils.inverse(other.ivar))
            new_ivar[np.invert(outmask_comb)] = 0
        else:
            new_ivar = None

        # RN2
        if self.rn2img is not None:
            new_rn2 = self.rn2img + other.rn2img
        else:
            new_rn2 = None

        # Files
        new_files = self.files + other.files

        # Instantiate
        new_sciImg = ScienceImage(self.spectrograph, self.det, self.par,
            newimg, new_ivar, self.bpm, rn2img=new_rn2, files=new_files)
        #TODO: KW properly handle adding the bits
        crmask_diff = new_sciImg.build_crmask()
        # crmask_eff assumes evertything masked in the outmask_comb is a CR in the individual images
        new_sciImg.crmask = crmask_diff | np.invert(outmask_comb)
        # Note that the following uses the saturation and mincounts held in
        # self.spectrograph.detector[self.det-1]
        new_sciImg.build_mask()

        return new_sciImg

    def __repr__(self):
        repr = '<{:s}: files={}'.format(self.__class__.__name__, self.files)
        # Image
        rdict = {}
        for attr in ['image', 'ivar', 'rn2img', 'crmask', 'mask']:
            if getattr(self, attr) is not None:
                rdict[attr] = True
            else:
                rdict[attr] = False
        repr += ' images={}'.format(rdict)
        repr = repr + '>'
        return repr


def build_from_file_list(spectrograph, det, par, bpm,
                   file_list, bias, pixel_flat, illum_flat=None,
                   sigma_clip=False, sigrej=None, maxiters=5):
    """
    Build a ScienceImage from a file list
    using a default set of process steps

    This will also generate the ivar, crmask, rn2img and mask

    Args:
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            Spectrograph used to take the data.
        det (:obj:`int`):
            The 1-indexed detector number to process.
        par (:class:`pypeit.par.pypeitpar.ProcessImagesPar`):
            Parameters that dictate the processing of the images.  See
            :class:`pypeit.par.pypeitpar.ProcessImagesPar` for the
            defaults.
        bpm (np.ndarray):
            Bad pixel mask.  Held in ImageMask
        file_list (list):
            List of files
        bias (np.ndarray or None):
            Bias image
        pixel_flat (np.ndarray):
            Flat image
        illum_flat (np.ndarray, optional):
            Illumination image
        sigrej (int or float, optional): Rejection threshold for sigma clipping.
             Code defaults to determining this automatically based on the numberr of images provided.
        maxiters (int, optional):

    Returns:
        ScienceImage:

    """
    # Process steps
    process_steps = procimg.init_process_steps(bias, par)
    process_steps += ['trim', 'apply_gain', 'orient']
    if (pixel_flat is not None) or (illum_flat is not None):
        process_steps += ['flatten']
    process_steps += ['extras']
    if par['cr_reject']:
        process_steps += ['crmask']

    combineImage = combineimage.CombineImage(spectrograph, det, par, file_list)
    pypeitImage = combineImage.run(process_steps, bias, bpm=bpm, pixel_flat=pixel_flat,
                                 illum_flat=illum_flat, sigma_clip=sigma_clip,
                                 sigrej=sigrej, maxiters=maxiters)

    # Instantiate
    slf = ScienceImage(spectrograph, det, par, pypeitImage.image, pypeitImage.ivar,
                                pypeitImage.bpm, rn2img=pypeitImage.rn2img,
                                crmask=pypeitImage.crmask, mask=pypeitImage.mask,
                                files=file_list)
    # Return
    return slf





"""
Implements the flat-field class.

.. include common links, assuming primary doc root is up one directory
.. include:: ../links.rst
"""
import os
import inspect
import numpy as np

from IPython import embed

from pypeit import msgs
from pypeit import ginga
from pypeit import masterframe

from pypeit.par import pypeitpar
from pypeit.images import calibrationimage
from pypeit.images import pypeitimage
from pypeit.core import flat
from pypeit.core import save
from pypeit.core import load
from pypeit.core import pixels
from pypeit.core import procimg


class FlatField(calibrationimage.CalibrationImage, masterframe.MasterFrame):
    """
    Builds pixel-level flat-field and the illumination flat-field.

    Args:
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            The `Spectrograph` instance that sets the instrument used to
            take the observations.  See usage by
            :class:`pypeit.processimages.ProcessImages` base class.
        par (:class:`pypeit.par.pypeitpar.FrameGroupPar`):
            The parameters used to type and process the flat frames.
        files (:obj:`list`, optional):
            The list of files to process.  Can be an empty list.
        det (:obj:`int`, optional):
            The 1-indexed detector number to process.
        master_key (:obj:`str`, optional):
            The string identifier for the instrument configuration.  See
            :class:`pypeit.masterframe.MasterFrame`.
        master_dir (:obj:`str`, optional):
            Path to master frames
        msbias (`numpy.ndarray`_, :obj:`str`, optional):
            Either an image with the bias to be subtracted, or a string
            providing the method to use for bias correction.
        msbpm (`numpy.ndarray`_, optional):
            Bad pixel mask image
        flatpar (:class:`pypeit.par.pypeitpar.FlatFieldPar`, optional):
            User-level parameters for constructing the flat-field
            corrections.  If None, the default parameters are used.
        tslits_dict (:obj:`dict`):
            The current slit traces; see
            :class:`pypeit.edgetrace.EdgeTraceSet`.
        tilts_dict (:obj:`dict`, optional):
            The current wavelength tilt traces; see
            :class:`pypeit.wavetilts.WaveTilts`.
        reuse_masters (:obj:`bool`, optional):
            Reuse already created master files from disk.

    Attributes:
        rawflatimg (PypeItImage):
        mspixelflat (ndarray):
            Normalized flat
        msillumflat (ndarray):
            Illumination flat
    """

    # Frame type is a class attribute
    frametype = 'pixelflat'
    master_type = 'Flat'

    @classmethod
    def from_master_file(cls, master_file, par=None):
        """
        Instantiate the class from a master file

        Args:
            master_file (str):
            par (:class:`pypeit.par.pypeitpar.PypeItPar`, optional):
                Full par set

        Returns:
            :class:`pypeit.flatfield.FlatField`:
                With the flat images loaded up

        """
        # Spectrograph
        spectrograph, extras = masterframe.items_from_master_file(master_file)
        head0 = extras[0]
        # Par
        if par is None:
            par = spectrograph.default_pypeit_par()
        # Master info
        master_dir = head0['MSTRDIR']
        master_key = head0['MSTRKEY']
        # Instantiate
        slf = cls(spectrograph, par['calibrations']['pixelflatframe'], master_dir=master_dir, master_key=master_key,
                  reuse_masters=True)
        # Load
        rawflatimg, slf.mspixelflat, slf.msillumflat = slf.load(ifile=master_file)
        # Convert rawflatimg to a PypeItImage
        slf.rawflatimg = pypeitimage.PypeItImage(rawflatimg)
        # Return
        return slf

    def __init__(self, spectrograph, par, files=None, det=1, master_key=None,
                 master_dir=None, reuse_masters=False, flatpar=None, msbias=None, msbpm=None,
                 tslits_dict=None, tilts_dict=None):

        # Image processing parameters
        if not isinstance(par, pypeitpar.FrameGroupPar):
            msgs.error('Must provide a FrameGroupPar instance as the parameters for FlatField.')
        self.par = par

        # Instantiate the base classes
        #   - Basic processing of the raw images
        calibrationimage.CalibrationImage.__init__(self, spectrograph, det, self.par['process'], files=files)
        #   - Construction and interface as a master frame
        masterframe.MasterFrame.__init__(self, self.master_type, master_dir=master_dir,
                                         master_key=master_key, reuse_masters=reuse_masters)

        # FieldFlattening parameters
        self.flatpar = pypeitpar.FlatFieldPar() if flatpar is None else flatpar

        # Input master frame data
        self.msbias = msbias
        self.msbpm = msbpm
        self.tslits_dict = tslits_dict
        self.tilts_dict = tilts_dict

        # Attributes unique to this Object
        self.rawflatimg = None      # Un-normalized pixel flat as a PypeItImage
        self.mspixelflat = None     # Normalized pixel flat
        self.msillumflat = None     # Slit illumination flat
        self.flat_model = None      # Model flat

        # Completed steps
        self.steps = []

        # Child-specific Internals
#        self.extrap_slit = None
#        self.msblaze = None
#        self.blazeext = None
#        self.slit_profiles = None
#        self.ntckx = None
#        self.ntcky = None

    @property
    def nslits(self):
        """
        Number of slits from the :attr:`tslits_dict` key=slit_left

        Returns:
            int: Number of slits

        """
        if self.tslits_dict is not None:
            return self.tslits_dict['slit_left'].shape[1]
        else:
            return 0

    def build_pixflat(self, trim=True, force=False):
        """
        Generate the flat image.

        Args:
            trim (:obj:`bool`, optional):
                Trim the image down to just the data section.
            force (:obj:`bool`, optional):
                Force the flat to be reconstructed if it already exists

        Returns:
            pypeitimage.PypeItImage:  The image with the unnormalized pixel-flat data.
        """
        if self.rawflatimg is None or force:
            # Process steps
            self.process_steps = procimg.init_process_steps(self.msbias, self.par['process'])
            ## JFH We never need untrimmed images. Why is this even an option?
            if trim:
                self.process_steps += ['trim']
            self.process_steps += ['apply_gain']
            self.process_steps += ['orient']
            # Turning this on leads to substantial edge-tracing problems when last tested
            #     JXP November 22, 2019
            #if self.par['cr_reject']:
            #    self.process_steps += ['crmask']
            self.steps.append(inspect.stack()[0][3])
            # Do it
            self.rawflatimg = super(FlatField, self).build_image(bias=self.msbias,
                                                                 bpm=self.msbpm,
                                                                 ignore_saturation=True)
        return self.rawflatimg

    # TODO Need to add functionality to use a different frame for the ilumination flat, e.g. a sky flat
    def run(self, debug=False, show=False, maskslits=None):
        """
        Generate normalized pixel and illumination flats

        Code flow::
           1.  Generate the pixelflat image (if necessary)
           2.  Prepare b-spline knot spacing
           3.  Loop on slits/orders
               a. Calculate the slit profile
               b. Normalize
               c. Save

        Args:
            debug (:obj:`bool`, optional):
                Run in debug mode.
            show (:obj:`bool`, optional):
                Show the results in the ginga viewer.
            maskslits (np.ndarray, optional):
               Array specifying whether a slit is good.
               True = bad

        Returns:
            `numpy.ndarray`_: Two arrays are returned, the normalized
            pixel flat data and the slit illumination correction data.
        """
        # Mask
        if maskslits is None:
            maskslits = np.zeros(self.nslits, dtype=bool)

        # Build the pixel flat (as needed)
        self.build_pixflat()

        # Prep tck (sets self.ntckx, self.ntcky)
        #self._prep_tck()

        # Setup
        self.mspixelflat = np.ones_like(self.rawflatimg.image)
        self.msillumflat = np.ones_like(self.rawflatimg.image)
        self.flat_model = np.zeros_like(self.rawflatimg.image)
        self.slitmask = pixels.tslits2mask(self.tslits_dict)

        final_tilts = np.zeros_like(self.rawflatimg.image)

        # If we are tweaking slits allocate the new aray to hold tweaked slit boundaries
        if self.flatpar['tweak_slits']:
            self.tslits_dict['slit_left_tweak'] = np.zeros_like(self.tslits_dict['slit_left'])
            self.tslits_dict['slit_righ_tweak'] = np.zeros_like(self.tslits_dict['slit_righ'])

        # Loop on slits
        for slit in range(self.nslits):
            # Is this a good slit??
            if maskslits[slit]:
                msgs.info('Skipping bad slit: {}'.format(slit))
                continue
            #
            msgs.info('Computing flat field image for slit: {:d}/{:d}'.format(slit,self.nslits-1))
            if self.msbpm is not None:
                inmask = np.invert(self.msbpm)
            else:
                inmask = np.ones_like(self.rawflatimg.image,dtype=bool)

            # Fit flats for a single slit
            this_tilts_dict = {'tilts':self.tilts_dict['tilts'],
                               'coeffs':self.tilts_dict['coeffs'][:,:,slit].copy(),
                               'slitcen':self.tilts_dict['slitcen'][:,slit].copy(),
                               'func2d':self.tilts_dict['func2d']}
            nonlinear_counts = self.spectrograph.nonlinear_counts(det=self.det)

            pixelflat, illumflat, flat_model, tilts_out, thismask_out, slit_left_out, \
                    slit_righ_out \
                            = flat.fit_flat(self.rawflatimg.image, this_tilts_dict, self.tslits_dict,
                                           slit, inmask=inmask, nonlinear_counts=nonlinear_counts,
                                           spec_samp_fine=self.flatpar['spec_samp_fine'],
                                           spec_samp_coarse=self.flatpar['spec_samp_coarse'],
                                           spat_samp=self.flatpar['spat_samp'],
                                           tweak_slits=self.flatpar['tweak_slits'],
                                           tweak_slits_thresh=self.flatpar['tweak_slits_thresh'],
                                           tweak_slits_maxfrac=self.flatpar['tweak_slits_maxfrac'],
                                           debug=debug)

            self.mspixelflat[thismask_out] = pixelflat[thismask_out]
            self.msillumflat[thismask_out] = illumflat[thismask_out]
            self.flat_model[thismask_out] = flat_model[thismask_out]

            # Did we tweak slit boundaries? If so, update the tslits_dict and the tilts_dict
            if self.flatpar['tweak_slits']:
                self.tslits_dict['slit_left'][:,slit] = slit_left_out
                self.tslits_dict['slit_righ'][:,slit] = slit_righ_out
                self.tslits_dict['slit_left_tweak'][:,slit] = slit_left_out
                self.tslits_dict['slit_righ_tweak'][:,slit] = slit_righ_out
                final_tilts[thismask_out] = tilts_out[thismask_out]

        # If we tweaked the slits update the tilts_dict
        if self.flatpar['tweak_slits']:
            self.tilts_dict['tilts'] = final_tilts

        if show:
            # Global skysub is the first step in a new extraction so clear the channels here
            self.show(slits=True, wcs_match = True)

        # If illumination flat fielding is turned off, set the illumflat to be None.
        if not self.flatpar['illumflatten']:
            msgs.warn('No illumination flat will be applied to your data (illumflatten=False).')
            self.msillumflat = None

        # Return
        return self.mspixelflat, self.msillumflat

    def show(self, slits=True, wcs_match=True):
        """
        Show all of the flat field products

        Args:
            slits (bool, optional):
            wcs_match (bool, optional):

        """
        viewer, ch = ginga.show_image(self.mspixelflat, chname='pixeflat', cuts=(0.9, 1.1),
                                      wcs_match=wcs_match, clear=True)
        viewer, ch = ginga.show_image(self.msillumflat, chname='illumflat', cuts=(0.9, 1.1),
                                      wcs_match=wcs_match)
        viewer, ch = ginga.show_image(self.rawflatimg, chname='flat', wcs_match=wcs_match)
        viewer, ch = ginga.show_image(self.flat_model, chname='flat_model', wcs_match=wcs_match)

        if slits:
            if self.tslits_dict is not None:
                slit_ids = [edgetrace.get_slitid(self.rawflatimg.shape,
                                                 self.tslits_dict['slit_left'],
                                                 self.tslits_dict['slit_righ'], ii)[0]
                                for ii in range(self.tslits_dict['slit_left'].shape[1])]
                ginga.show_slits(viewer, ch, self.tslits_dict['slit_left'],
                                 self.tslits_dict['slit_righ'], slit_ids)

    def save(self, outfile=None, overwrite=True):
        """
        Save the flat-field master data to a FITS file

        Extensions are:
            RAWFLAT
            PIXELFLAT
            ILLUMFLAT
            MODEL

        Args:
            outfile (:obj:`str`, optional):
                Name for the output file.  Defaults to
                :attr:`file_path`.
            overwrite (:obj:`bool`, optional):
                Overwrite any existing file.
        """
        _outfile = self.master_file_path if outfile is None else outfile
        # Check if it exists
        if os.path.exists(_outfile) and not overwrite:
            msgs.warn('Master file exists: {0}'.format(_outfile) + msgs.newline()
                      + 'Set overwrite=True to overwrite it.')
            return

        # Setup the items
        hdr = self.build_master_header(steps=self.steps, raw_files=self.file_list)
        data = [self.rawflatimg.image, self.mspixelflat, self.msillumflat, self.flat_model]
        extnames = ['RAWFLAT', 'PIXELFLAT', 'ILLUMFLAT', 'MODEL']

        # Save to a multi-extension FITS
        save.write_fits(hdr, data, _outfile, extnames=extnames)
        msgs.info('Master frame written to {0}'.format(_outfile))

    def load(self, ifile=None):
        """
        Load the flat-field data from a save master frame.

        Args:
            ifile (:obj:`str`, optional):
                Name of the master frame file.  Defaults to
                :attr:`file_path`.

        Returns:
            tuple: Returns three `numpy.ndarray`_ objects with the raw
            flat-field image, the normalized pixel flat, and the
            illumination flat.
        """
        # Check on whether to reuse and whether the file exists
        master_file = self.chk_load_master(ifile)
        if master_file is None:
            return None, None, None
        # Load
        ext = ['RAWFLAT', 'PIXELFLAT', 'ILLUMFLAT', 'MODEL']
        self.rawflatimg, self.mspixelflat, self.msillumflat, self.flat_model, head0 = load.load_multiext_fits(master_file, ext)
        # Return
        return self.rawflatimg, self.mspixelflat, self.msillumflat




"""
Module to run tests on SpecObj
"""
import numpy as np
import sys
import os
import pytest

#import pypeit

from astropy.table import Table

from pypeit import specobj
from pypeit import msgs

def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)


def test_init():
    sobj = specobj.SpecObj('MultiSlit', 1, slitid=0)
    # Test
    assert sobj.PYPELINE == 'MultiSlit'
    assert sobj['PYPELINE'] == 'MultiSlit'
    assert sobj.NAME == 'SPAT-----SLIT0000-DET01'
    assert len(sobj._data.keys()) == 0


def test_assignment():
    sobj = specobj.SpecObj('MultiSlit', 1, slitid=0)
    #
    sobj.PYPELINE = 'Blah'
    #
    with pytest.raises(OSError):
        sobj.PYPELINE = 2
    #
    sobj.SPAT_PIXPOS = 523.0
    sobj.PYPELINE = 'MultiSlit'
    sobj.set_name()
    assert sobj.NAME == 'SPAT0523-SLIT0000-DET01'


def test_data():
    sobj = specobj.SpecObj('MultiSlit', 1, slitid=0)
    #
    sobj['BOX_WAVE'] = np.arange(100).astype(float)
    sobj['BOX_COUNTS'] = np.ones_like(sobj.BOX_WAVE)
    sobj['TRACE_SPAT'] = np.arange(100) * 2.
    # Test
    assert isinstance(sobj._data, Table)
    assert 'TRACE_SPAT' in sobj._data.keys()
    assert 'SLITID' in sobj._data.meta.keys()

def test_io():
    sobj = specobj.SpecObj('MultiSlit', 1, slitid=0)
    #
    sobj['BOX_WAVE'] = np.arange(100).astype(float)
    sobj['BOX_COUNTS'] = np.ones_like(sobj.BOX_WAVE)
    sobj['TRACE_SPAT'] = np.arange(100) * 2.

    # Write table
    sobj._data.write(data_path('tmp.fits'), overwrite=True)
    tbl = Table.read(data_path('tmp.fits'))
    sobj2 = specobj.SpecObj.from_table(tbl)
    #
    assert isinstance(sobj2, specobj.SpecObj)



#!/usr/bin/env python
#
# See top-level LICENSE file for Copyright information
#
# -*- coding: utf-8 -*-
"""
This script enables the viewing of a processed FITS file
with extras.  Run above the Science/ folder.
"""
import argparse
import os

import numpy as np

from IPython import embed

from astropy.io import fits
from astropy.stats import sigma_clipped_stats

from pypeit import msgs
from pypeit import ginga
from pypeit import edgetrace
from pypeit import specobjs
from pypeit.core import pixels
from pypeit.core.parse import get_dnum
from pypeit.images.maskimage import ImageBitMask
from pypeit.masterframe import MasterFrame


def parser(options=None):
    parser = argparse.ArgumentParser(description='Display sky subtracted, spec2d image in a '
                                                 'Ginga viewer.  Run above the Science/ folder',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('file', type = str, default = None, help = 'PYPIT spec2d file')
    parser.add_argument('--list', default=False, help='List the extensions only?',
                        action='store_true')
    parser.add_argument('--det', default=1, type=int, help='Detector number')
    parser.add_argument('--showmask', default=False, help='Overplot masked pixels',
                        action='store_true')
    parser.add_argument('--removetrace', default=False, help="Do not overplot traces in the skysub, "
                                                             "sky_resid and resid channels",
                        action = "store_true")
    parser.add_argument('--embed', default=False, help='Upon completion embed in ipython shell',
                        action='store_true')

    return parser.parse_args() if options is None else parser.parse_args(options)


def show_trace(specobjs, det, viewer, ch):

    if specobjs is None:
        return
    in_det = np.where(specobjs.DET == det)[0]
    for kk in in_det:
        trace = specobjs[kk]['TRACE_SPAT']
        obj_id = specobjs[kk].NAME
        ginga.show_trace(viewer, ch, trace, obj_id, color='orange') #hdu.name)


def main(args):

    # List only?
    hdu = fits.open(args.file)
    head0 = hdu[0].header
    if args.list:
        hdu.info()
        return

    # Setup for PYPIT imports
    msgs.reset(verbosity=2)

    # Init
    # TODO: get_dnum needs to be deprecated...
    sdet = get_dnum(args.det, prefix=False)

    # One detector, sky sub for now
    names = [hdu[i].name for i in range(len(hdu))]

    try:
        exten = names.index('DET{:s}-PROCESSED'.format(sdet))
    except:  # Backwards compatability
        msgs.error('Requested detector {:s} was not processed.\n'
                   'Maybe you chose the wrong one to view?\n'
                   'Set with --det= or check file contents with --list'.format(sdet))
    sciimg = hdu[exten].data
    try:
        exten = names.index('DET{:s}-SKY'.format(sdet))
    except:  # Backwards compatability
        msgs.error('Requested detector {:s} has no sky model.\n'
                   'Maybe you chose the wrong one to view?\n'
                   'Set with --det= or check file contents with --list'.format(sdet))
    skymodel = hdu[exten].data
    try:
        exten = names.index('DET{:s}-MASK'.format(sdet))
    except ValueError:  # Backwards compatability
        msgs.error('Requested detector {:s} has no bit mask.\n'
                   'Maybe you chose the wrong one to view?\n'
                   'Set with --det= or check file contents with --list'.format(sdet))
    mask = hdu[exten].data
    try:
        exten = names.index('DET{:s}-IVARMODEL'.format(sdet))
    except ValueError:  # Backwards compatability
        msgs.error('Requested detector {:s} has no IVARMODEL.\n'
                   'Maybe you chose the wrong one to view?\n' +
                   'Set with --det= or check file contents with --list'.format(sdet))
    ivarmodel = hdu[exten].data
    # Read in the object model for residual map
    try:
        exten = names.index('DET{:s}-OBJ'.format(sdet))
    except ValueError:  # Backwards compatability
        msgs.error('Requested detector {:s} has no object model.\n'
                   'Maybe you chose the wrong one to view?\n' +
                   'Set with --det= or check file contents with --list'.format(sdet))
    objmodel = hdu[exten].data
    # Get waveimg
    mdir = head0['PYPMFDIR']
    if not os.path.exists(mdir):
        mdir_base = os.path.join(os.getcwd(), os.path.basename(mdir))
        msgs.warn('Master file dir: {0} does not exist. Using {1}'.format(mdir, mdir_base))
        mdir=mdir_base

    trace_key = '{0}_{1:02d}'.format(head0['TRACMKEY'], args.det)
    trc_file = '{0}.gz'.format(os.path.join(mdir,
                                            MasterFrame.construct_file_name('Edges', trace_key)))

    wave_key = '{0}_{1:02d}'.format(head0['ARCMKEY'], args.det)
    waveimg = os.path.join(mdir, MasterFrame.construct_file_name('Wave', wave_key))

    tslits_dict = edgetrace.EdgeTraceSet.from_file(trc_file).convert_to_tslits_dict()
    slitmask = pixels.tslits2mask(tslits_dict)
    shape = (tslits_dict['nspec'], tslits_dict['nspat'])
    slit_ids = [edgetrace.get_slitid(shape, tslits_dict['slit_left'], tslits_dict['slit_righ'],
                                     ii)[0] for ii in range(tslits_dict['slit_left'].shape[1])]
    # Show the bitmask?
    mask_in = mask if args.showmask else None

    # Object traces from spec1d file
    spec1d_file = args.file.replace('spec2d', 'spec1d')

    if os.path.isfile(spec1d_file):
        sobjs = specobjs.SpecObjs.from_fitsfile(spec1d_file)
    else:
        sobjs = None
        msgs.warn('Could not find spec1d file: {:s}'.format(spec1d_file) + msgs.newline() +
                  '                          No objects were extracted.')

    # Unpack the bitmask
    bitMask = ImageBitMask()
    bpm, crmask, satmask, minmask, offslitmask, nanmask, ivar0mask, ivarnanmask, extractmask \
            = bitMask.unpack(mask)

    # Now show each image to a separate channel

    # SCIIMG
    image = sciimg # Raw science image
    (mean, med, sigma) = sigma_clipped_stats(image[mask == 0], sigma_lower=5.0, sigma_upper=5.0)
    cut_min = mean - 1.0 * sigma
    cut_max = mean + 4.0 * sigma
    chname_skysub='sciimg-det{:s}'.format(sdet)
    # Clear all channels at the beginning
    viewer, ch = ginga.show_image(image, chname=chname_skysub, waveimg=waveimg, bitmask=mask_in, clear=True)
                                  #, cuts=(cut_min, cut_max), wcs_match=True)
    if sobjs is not None:
        show_trace(sobjs, args.det, viewer, ch)
    ginga.show_slits(viewer, ch, tslits_dict['slit_left'], tslits_dict['slit_righ'], slit_ids)
                     #, args.det)

    # SKYSUB
    image = (sciimg - skymodel) * (mask == 0)  # sky subtracted image
    (mean, med, sigma) = sigma_clipped_stats(image[mask == 0], sigma_lower=5.0, sigma_upper=5.0)
    cut_min = mean - 1.0 * sigma
    cut_max = mean + 4.0 * sigma
    chname_skysub='skysub-det{:s}'.format(sdet)
    # Clear all channels at the beginning
    # TODO: JFH For some reason Ginga crashes when I try to put cuts in here.
    viewer, ch = ginga.show_image(image, chname=chname_skysub, waveimg=waveimg,
                                  bitmask=mask_in) #, cuts=(cut_min, cut_max),wcs_match=True)
    if not args.removetrace and sobjs is not None:
            show_trace(sobjs, args.det, viewer, ch)
    ginga.show_slits(viewer, ch, tslits_dict['slit_left'], tslits_dict['slit_righ'], slit_ids)


    # SKRESIDS
    chname_skyresids = 'sky_resid-det{:s}'.format(sdet)
    image = (sciimg - skymodel) * np.sqrt(ivarmodel) * (mask == 0)  # sky residual map
    viewer, ch = ginga.show_image(image, chname_skyresids, waveimg=waveimg,
                                  cuts=(-5.0, 5.0), bitmask = mask_in) #,wcs_match=True)
    if not args.removetrace and sobjs is not None:
            show_trace(sobjs, args.det, viewer, ch)
    ginga.show_slits(viewer, ch, tslits_dict['slit_left'], tslits_dict['slit_righ'], slit_ids)

    # RESIDS
    chname_resids = 'resid-det{:s}'.format(sdet)
    # full model residual map
    image = (sciimg - skymodel - objmodel) * np.sqrt(ivarmodel) * (mask == 0)
    viewer, ch = ginga.show_image(image, chname=chname_resids, waveimg=waveimg,
                                  cuts = (-5.0, 5.0), bitmask = mask_in)
    if not args.removetrace and sobjs is not None:
            show_trace(sobjs, args.det, viewer, ch)
    ginga.show_slits(viewer, ch, tslits_dict['slit_left'], tslits_dict['slit_righ'], slit_ids)


    # After displaying all the images sync up the images with WCS_MATCH
    shell = viewer.shell()
    out = shell.start_global_plugin('WCSMatch')
    out = shell.call_global_plugin_method('WCSMatch', 'set_reference_channel', [chname_resids], {})

    if args.embed:
        embed()

        # Playing with some mask stuff
        #out = shell.start_operation('TVMask')
        #maskfile = '/Users/joe/python/PypeIt-development-suite/REDUX_OUT/Shane_Kast_blue/600_4310_d55/shane_kast_blue_setup_A/crmask.fits'
        #out = shell.call_local_plugin_method(chname_resids, 'TVMask', 'load_file', [maskfile], {})





""" Lightweight object to load and hold the RawImage and several other key
items for its processing."""

import numpy as np

from IPython import embed


class RawImage(object):
    """
    Class to load and hold a raw image

    Args:
        filename (:obj:`str` or None):
            Filename
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            Spectrograph used to take the data.
        det (:obj:`int`):
            The 1-indexed detector number to process.

    Attributes:
        raw_image (np.ndarray):
            Raw image as a numpy array
        rawdatasec_img (np.ndarray):
            Holds the datasec_img which specifies the amp for each pixel in the
            raw image
        oscansec_img (np.ndarray):
            Holds the oscansec_img once loaded.  This is in the raw frame
        hdu (fits.HDUList):
            HDUList of the file
        exptime (float):
            Exposure time
    """
    def __init__(self, filename, spectrograph, det):

        # Init me
        self.filename = filename
        self.spectrograph = spectrograph
        self.det = det

        # Load the raw image and the other items of interest
        self.raw_image, self.hdu, self.exptime, self.rawdatasec_img, self.oscansec_img = self.spectrograph.get_rawimage(
            self.filename, self.det)

    def __repr__(self):
        return ('<{:s}: file={}> spectrograph={}'.format(
            self.__class__.__name__, self.filename, self.spectrograph.spectrograph))




"""
Module to run tests on arsave
"""
import os

import numpy as np

import pytest

from pypeit import msgs
from pypeit.par.util import make_pypeit_file
from pypeit import pypeitsetup
from pypeit.pypeit import PypeIt

def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)


def test_initialization():
    """ Load input PYPIT file
    """
    # Generate a PYPIT file
    pypit_file = data_path('test.pypeit')
    make_pypeit_file(pypit_file, 'shane_kast_blue', [data_path('b*fits.gz')], setup_mode=True)

    # Perform the setup
    setup = pypeitsetup.PypeItSetup.from_pypeit_file(pypit_file)
    par, spectrograph, fitstbl = setup.run(sort_dir=data_path(''))

    # Test
    assert spectrograph.spectrograph == 'shane_kast_blue'
    assert len(fitstbl) == 2
    assert par['calibrations']['arcframe']['number'] == 1

    # Clean-up
    os.remove(data_path('test.calib'))
    os.remove(data_path('test.pypeit'))

def test_select_detectors():
    # Generate a PYPIT file
    pypit_file = data_path('test.pypeit')
    make_pypeit_file(pypit_file, 'shane_kast_blue', [data_path('b*fits.gz')], setup_mode=True)

    # Perform the setup
    setup = pypeitsetup.PypeItSetup.from_pypeit_file(pypit_file)
    par, spectrograph, fitstbl = setup.run(sort_dir=data_path(''))

    assert PypeIt.select_detectors(detnum=par['rdx']['detnum'], ndet=spectrograph.ndet) == [1], \
            'Incorrect detectors selected.'

    # Clean-up
    os.remove(data_path('test.calib'))
    os.remove(data_path('test.pypeit'))

    assert np.array_equal(PypeIt.select_detectors(), [1]), 'Incorrect detectors selected.'
    assert np.array_equal(PypeIt.select_detectors(detnum=3, ndet=5), [3]), \
            'Incorrect detectors selected.'
    assert np.array_equal(PypeIt.select_detectors(ndet=5), [1,2,3,4,5]), \
            'Incorrect detectors selected.'
    assert np.array_equal(PypeIt.select_detectors(detnum=[1,3]), [1,3]), \
            'Incorrect detectors selected.'




# encoding: utf-8
"""
Define the telescopes parameters used by Pypit.
"""
from pypeit.par.pypeitpar import TelescopePar

#TODO: Remove 'Par' from class name?

class GeminiNTelescopePar(TelescopePar):
    def __init__(self):
        super(GeminiNTelescopePar, self).__init__(name='GEMINI-N',
                                               longitude=155.47833,
                                               latitude=19.82833,
                                               elevation=4160.0)

class KeckTelescopePar(TelescopePar):
    def __init__(self):
        super(KeckTelescopePar, self).__init__(name='KECK',
                                               longitude=155.47833,
                                               latitude=19.82833,
                                               elevation=4160.0,
                                               fratio=15,
                                               diameter=10)

class MagellanTelescopePar(TelescopePar):
    def __init__(self):
        super(MagellanTelescopePar, self).__init__(name='MAGELLAN',
                                               longitude=70.6858,
                                               latitude=-29.0283,
                                               elevation=2516.0,
                                               diameter=6.5)

class ShaneTelescopePar(TelescopePar):
    def __init__(self):
        super(ShaneTelescopePar, self).__init__(name='SHANE',
                                                longitude=121.6428,
                                                latitude=37.3413889,
                                                elevation=1283.0)
                                
class WHTTelescopePar(TelescopePar):
    def __init__(self):
        super(WHTTelescopePar, self).__init__(name='WHT',
                                              longitude=17.8947,
                                              latitude=26.7636,
                                              elevation=2396.0)
                                
class APFTelescopePar(TelescopePar):
    def __init__(self):
        super(APFTelescopePar, self).__init__(name='APF',
                                              longitude=121.642778,
                                              latitude=37.34138889,
                                              elevation=1283.0)
                                
class TNGTelescopePar(TelescopePar):
    def __init__(self):
        super(TNGTelescopePar, self).__init__(name='TNG',
                                              longitude=17.88906,
                                              latitude=28.754,
                                              elevation=2387.2)

class VLTTelescopePar(TelescopePar):
    def __init__(self):
        super(VLTTelescopePar, self).__init__(name='VLT',
                                               longitude=70.404830556,
                                               latitude=-24.6271666666,
                                               elevation=2635.43)
class GeminiSTelescopePar(TelescopePar):
    def __init__(self):
        super(GeminiSTelescopePar, self).__init__(name='GEMINI-S',
                                              longitude=70.8062,              # Longitude of the telescope (NOTE: West should correspond to positive longitudes)
                                              latitude=-30.1691,              # Latitude of the telescope
                                              elevation=2200.0)               # Elevation of the telescope (in m)

class LBTTelescopePar(TelescopePar):
    def __init__(self):
        super(LBTTelescopePar, self).__init__(name='LBT',
                                               longitude=109.889064,
                                               latitude=32.701308,
                                               elevation=3221.0)

class MMTTelescopePar(TelescopePar):
    def __init__(self):
        super(MMTTelescopePar, self).__init__(name='MMT',
                                               longitude=110.885,
                                               latitude=31.6883,
                                               elevation=2616.0)



"""
Module to run tests on BiasFrame class
Requires files in Development suite and an Environmental variable
"""
import os

import pytest
import glob
import numpy as np

from pypeit import biasframe
from pypeit.tests.tstutils import dev_suite_required
from pypeit.spectrographs.util import load_spectrograph

shane_kast_blue = load_spectrograph('shane_kast_blue')

def data_root():
    return os.path.join(os.path.dirname(__file__), 'files')

@pytest.fixture
@dev_suite_required
def deimos_flat_files():
    # Longslit in dets 3,7
    deimos_flat_files = [os.path.join(os.getenv('PYPEIT_DEV'),
                                      '/RAW_DATA/keck_deimos/830G_L/', ifile) \
                            for ifile in ['d0914_0014.fits', 'd0914_0015.fits'] ]
    assert len(deimos_flat_files) == 2
    return deimos_flat_files

@pytest.fixture
@dev_suite_required
def kast_blue_bias_files():
    kast_blue_bias_files = glob.glob(os.path.join(os.getenv('PYPEIT_DEV'), 'RAW_DATA',
                                                  'shane_kast_blue', '600_4310_d55', 'b1?.fits*'))
    return kast_blue_bias_files

@dev_suite_required
def test_instantiate(kast_blue_bias_files):
    # Empty
    bias_frame0 = biasframe.BiasFrame(shane_kast_blue)
    assert bias_frame0.nfiles == 0
    #
    bias_frame1 = biasframe.BiasFrame(shane_kast_blue, files=kast_blue_bias_files)
    assert bias_frame1.nfiles == 10

@dev_suite_required
def test_process(kast_blue_bias_files):
    # Instantiate
    bias_frame = biasframe.BiasFrame(shane_kast_blue, files=kast_blue_bias_files)
    # Run
    bias_img = bias_frame.build_image()
    assert isinstance(bias_img.image, np.ndarray)
    assert isinstance(bias_frame.pypeitImage.image, np.ndarray)
    #assert bias_frame.steps[-1] == 'combine'


@dev_suite_required
def test_io(kast_blue_bias_files):
    # Instantiate
    bias_frame = biasframe.BiasFrame(shane_kast_blue, files=kast_blue_bias_files,
                                     master_dir=data_root(), master_key='A_01_1', 
                                     reuse_masters=True)
    # In case of previous test failure
    if os.path.isfile(bias_frame.master_file_path):
        os.remove(bias_frame.master_file_path)
    # Run
    bias_frame.build_image()
    # Save as a master frame
    bias_frame.save()
    assert os.path.isfile(bias_frame.master_file_path), 'Error writing MasterBias'
    # Load master frame
    pypeitImage = bias_frame.load()
    assert np.array_equal(pypeitImage.image, bias_frame.pypeitImage.image)
    # Instantiate from master frame
    bias_frame2 = biasframe.BiasFrame.from_master_file(bias_frame.master_file_path)
    assert np.array_equal(pypeitImage.image, bias_frame2.pypeitImage.image)
    # Clean up
    os.remove(bias_frame.master_file_path)


@dev_suite_required
def test_run_and_master(kast_blue_bias_files):
    # Instantiate
    bias_frame = biasframe.BiasFrame(shane_kast_blue, files=kast_blue_bias_files,
                                     master_key='A_1_01', master_dir=data_root())
    assert bias_frame.frametype == 'bias'
    # In case of previous test failure
    if os.path.isfile(bias_frame.master_file_path):
        os.remove(bias_frame.master_file_path)

    # Run
    msbias = bias_frame.build_image()
    #assert bias_frame.steps[-1] == 'combine'
    # Save it
    bias_frame.save()

    # Run with reuse (should simply load the file)
    bias_frame2 = biasframe.BiasFrame(shane_kast_blue, master_key='A_1_01',
                                      master_dir=data_root(), reuse_masters=True)
    bias2 = bias_frame2.load()
    assert isinstance(bias2.image, np.ndarray)
    assert len(bias_frame2.process_steps) == 3
    assert np.array_equal(bias2.image, bias_frame.pypeitImage.image)

    # Clean up
    os.remove(bias_frame.master_file_path)

# Should probably test overscan




"""
Module for the SpecObjs and SpecObj classes

.. include common links, assuming primary doc root is up one directory
.. include:: ../links.rst
"""
import os
import re

import numpy as np

from astropy.units import Quantity
from astropy.io import fits

from pypeit import msgs
from pypeit.core import save
from pypeit import specobj
from pypeit.io import initialize_header
from pypeit.spectrographs.util import load_spectrograph

from IPython import embed

class SpecObjs(object):
    """
    Object to hold a set of SpecObj objects

    Note that this class overloads:

        - ``__getitem__`` to allow one to pull an attribute or a portion
          of the SpecObjs list
        - ``__setattr__`` to force a custom assignment method
        - ``__getattr__`` to generate an array of attribute 'k' from the
          specobjs.

    Args:
        specobjs (ndarray or list, optional):  One or more SpecObj objects

    Attributes:
        summary (astropy.table.Table):
    """
    @classmethod
    def from_fitsfile(cls, fits_file):
        """
        Instantiate from a FITS file

        Also tag on the Header

        Args:
            fits_file (str):

        Returns:
            specobsj.SpecObjs

        """
        # HDUList
        hdul = fits.open(fits_file)
        nhdu = len(hdul)
        # Init
        slf = cls()
        # Add on the header
        slf.header = hdul[0].header
        # Loop on em
        for kk in range(1,nhdu):
            tbl = fits.connect.read_table_fits(hdul, hdu=kk)
            sobj = specobj.SpecObj.from_table(tbl)
            slf.add_sobj(sobj)

        # JFH I'm commenting this out below. I prefer to just directly write out attributes and reinstantiate them
        # from files. Doing things like this just leads to errors
        # since it is not in touch with the code that actually determined what these attributes should be.

        # PYPELINE specific
        #if slf[0].PYPELINE == 'Echelle':
        #    # Set ech_objid
        #    uni_frac = np.unique(slf.ECH_FRACPOS)
        #    for ii, ufrac in enumerate(uni_frac):
        #        idx = np.isclose(slf.ECH_FRACPOS, ufrac)
        #        slf[idx].ECH_OBJID = ii
        #    # Set ech_orderindx
        #    uni_order = np.unique(slf.ECH_ORDER)
        #    for ii, uorder in enumerate(uni_order):
        #        idx = slf.ECH_ORDER == uorder
        #        slf[idx].ech_orderindx = ii

        # Return
        return slf

    def __init__(self, specobjs=None, header=None):

        # Only one attribute is allowed for this Object -- specobjs
        if specobjs is None:
            self.specobjs = np.array([])
        else:
            if isinstance(specobjs, (list, np.ndarray)):
                specobjs = np.array(specobjs)
            self.specobjs = specobjs

        self.header = header if header is not None else None

        # Turn off attributes from here
        #   Anything else set will be on the individual specobj objects in the specobjs array
        self.__initialised = True

    def __setattr__(self, item, value):
        """
        Define a custom assignment method.

        Args:
            item (str):
                Item to set
            value (object):
                Value of the item

        """
        if not '_SpecObjs__initialised' in self.__dict__:  # this test allows attributes to be set in the __init__ method
            return dict.__setattr__(self, item, value)
        elif item in self.__dict__:  # any normal attributes are handled normally
            dict.__setattr__(self, item, value)
        else:
            # Special handling when the input is an array/list and the length matches that of the slice
            if isinstance(value, (list, np.ndarray)):
                if len(self.specobjs) == len(value):  # Assume these are to be paired up
                    for kk, specobj in enumerate(self.specobjs):
                        setattr(specobj, item, value[kk])
                    return
            #
            for specobj in self.specobjs:
                setattr(specobj, item, value)

    @property
    def nobj(self):
        """
        Return the number of SpecObj objects

        Returns:
            int

        """
        return len(self.specobjs)

    def unpack_object(self, ret_flam=False):
        """
        Utility function to unpack the sobjs for one object and
        return various numpy arrays describing the spectrum and meta
        data. The user needs to already have trimmed the Specobjs to
        the relevant indices for the object.

        Args:
           ret_flam (:obj:`bool`, optional):
              If True return the FLAM, otherwise return COUNTS.

        Returns:
            tuple: Returns the following where all numpy arrays
            returned have shape (nspec, norders) for Echelle data and
            (nspec,) for Multislit data.

                - wave (`numpy.ndarray`_): Wavelength grids
                - flux (`numpy.ndarray`_): Flambda or counts
                - flux_ivar (`numpy.ndarray`_): Inverse variance (of
                  Flambda or counts)
                - flux_gpm (`numpy.ndarray`_): Good pixel mask.
                  True=Good
                - meta_spec (dict:) Dictionary containing meta data.
                  The keys are defined by
                  spectrograph.header_cards_from_spec()
                - header (astropy.io.header object): header from
                  spec1d file
    """

        # Read in the spec1d file
        norddet = self.nobj
        if ret_flam:
            # TODO Should nspec be an attribute of specobj?
            nspec = self[0].OPT_FLAM.size
        else:
            nspec = self[0].OPT_COUNTS.size
        # Allocate arrays and unpack spectrum
        wave = np.zeros((nspec, norddet))
        flux = np.zeros((nspec, norddet))
        flux_ivar = np.zeros((nspec, norddet))
        flux_gpm = np.zeros((nspec, norddet), dtype=bool)
        detector = np.zeros(norddet, dtype=int)
        ech_orders = np.zeros(norddet, dtype=int)
        # TODO make the extraction that is desired OPT vs BOX and optional input variable.
        for iorddet in range(norddet):
            wave[:, iorddet] = self[iorddet].OPT_WAVE
            flux_gpm[:, iorddet] = self[iorddet].OPT_MASK
            detector[iorddet] = self[iorddet].DET
            if self[0].PYPELINE == 'Echelle':
                ech_orders[iorddet] = self[iorddet].ECH_ORDER
            if ret_flam:
                flux[:, iorddet] = self[iorddet].OPT_FLAM
                flux_ivar[:, iorddet] = self[iorddet].OPT_FLAM_IVAR
            else:
                flux[:, iorddet] = self[iorddet].OPT_COUNTS
                flux_ivar[:, iorddet] = self[iorddet].OPT_COUNTS_IVAR

        # Populate meta data
        # TODO Remove this hack is it needed? If PYP_SPEC is always written then it is not.
        try:
            spectrograph = load_spectrograph(self.header['PYP_SPEC'])
        except:
            # TODO JFH  This is a hack until a generic spectrograph is implemented.
            spectrograph = load_spectrograph('shane_kast_blue')

        meta_spec = {}
        core_keys = spectrograph.header_cards_for_spec()
        for key in core_keys:
            try:
                meta_spec[key.upper()] = self.header[key.upper()]
            except KeyError:
                msgs.warn('Core meta data is missing from the specobjs header ')
                pass
        # Add the pyp spec.
        # TODO JFH: Make this an atribute of the specobj by default.
        meta_spec['PYP_SPEC'] = self.header['PYP_SPEC']
        meta_spec['PYPELINE'] = self[0].PYPELINE
        meta_spec['DET'] = detector
        if self[0].PYPELINE == 'MultiSlit' and self.nobj == 1:
            meta_spec['ECH_ORDERS'] = None
            return wave.reshape(nspec), flux.reshape(nspec), flux_ivar.reshape(nspec), \
                   flux_gpm.reshape(nspec), meta_spec, self.header
        else:
            meta_spec['ECH_ORDERS'] = ech_orders
            return wave, flux, flux_ivar, flux_gpm, meta_spec, self.header






    def get_std(self, multi_spec_det=None):
        """
        Return the standard star from this Specobjs. For MultiSlit this
        will be a single specobj in SpecObjs container, for Echelle it
        will be the standard for all the orders.

        Args:
            multi_spec_det (list):
                If there are multiple detectors arranged in the spectral direction, return the sobjs for
                the standard on each detector.

        Returns:
            SpecObj or SpecObjs

        """
        # Is this MultiSlit or Echelle
        pypeline = (self.PYPELINE)[0]
        if 'MultiSlit' in pypeline:
            # Have to do a loop to extract the counts for all objects
            SNR = np.median(self.OPT_COUNTS*np.sqrt(self.OPT_COUNTS_IVAR), axis=1)
            # For multiple detectors grab the requested detectors
            if multi_spec_det is not None:
                sobjs_std = SpecObjs(header=self.header)
                # Now append the maximum S/N object on each detector
                for idet in multi_spec_det:
                    this_det = self.DET == idet
                    istd = SNR[this_det].argmax()
                    sobjs_std.add_sobj(self[this_det][istd])
            else: # For normal multislit take the brightest object
                istd = SNR.argmax()
                # Return
                sobjs_std = SpecObjs(specobjs=[self[istd]], header=self.header)
            sobjs_std.header = self.header
            return sobjs_std
        elif 'Echelle' in pypeline:
            uni_objid = np.unique(self.ECH_FRACPOS)  # A little risky using floats
            uni_order = np.unique(self.ECH_ORDER)
            nobj = len(uni_objid)
            norders = len(uni_order)
            # Build up S/N
            SNR = np.zeros((norders, nobj))
            for iobj in range(nobj):
                for iord in range(norders):
                    ind = (self.ECH_FRACPOS == uni_objid[iobj]) & (self.ECH_ORDER == uni_order[iord])
                    spec = self[ind]
                    SNR[iord, iobj] = np.median(spec[0].OPT_COUNTS*np.sqrt(spec[0].OPT_COUNTS_IVAR))
            # Maximize S/N
            SNR_all = np.sqrt(np.sum(SNR**2,axis=0))
            objid_std = uni_objid[SNR_all.argmax()]
            # Finish
            indx = self.ECH_FRACPOS == objid_std
            # Return
            sobjs_std = SpecObjs(specobjs=self[indx], header=self.header)
            sobjs_std.header = self.header
            return sobjs_std
        else:
            msgs.error('Unknown pypeline')

    def append_neg(self, sobjs_neg):
        """
        Append negative objects and change the sign of their objids for IR reductions

        Args:
            sobjs_neg (SpecObjs):

        """
        if sobjs_neg.nobj == 0:
            msgs.warn("No negative objects found...")
            return
        # Assign the sign and the objids
        sobjs_neg.sign = -1.0
        if sobjs_neg[0].PYPELINE == 'Echelle':
            sobjs_neg.ECH_OBJID = -sobjs_neg.ECH_OBJID
            sobjs_neg.OBJID = -sobjs_neg.OBJID
        elif sobjs_neg[0].PYPELINE == 'MultiSlit':
            sobjs_neg.OBJID = -sobjs_neg.OBJID
        else:
            msgs.error("Should not get here")
        self.add_sobj(sobjs_neg)

        # Sort objects according to their spatial location. Necessary for the extraction to properly work
        if self.nobj > 0:
            self.specobjs = self.specobjs[np.argsort(self.SPAT_PIXPOS)]

    def purge_neg(self):
        """
        Purge negative objects from specobjs for IR reductions

        """
        # Assign the sign and the objids
        if self.nobj > 0:
            if self[0].PYPELINE == 'Echelle':
                index = self.ECH_OBJID < 0
            elif self[0].PYPELINE == 'MultiSlit':
                index = self.OBJID < 0
            else:
                msgs.error("Should not get here")
            self.remove_sobj(index)


    def slitorder_indices(self, slitorder):
        """
        Return the set of indices matching the input slit/order
        """
        if self[0].PYPELINE == 'Echelle':
            indx = self.ECH_ORDERINDX == slitorder
        elif self[0].PYPELINE == 'MultiSlit':
            indx = self.SLITID == slitorder
        else:
            msgs.error("Should not get here")
        #
        return indx

    def name_indices(self, name):
        """
        Return the set of indices matching the input slit/order

        Parameters
        ----------
        name: str
           The name of the object

        Returns
        -------
        indx: array-like, shape (nobj,)
           Array of indices with the corresponding name.

        """
        if self[0].PYPELINE == 'Echelle':
            indx = self.ECH_NAME == name
        elif self[0].PYPELINE == 'MultiSlit':
            indx = self.NAME == name
        else:
            msgs.error("Should not get here")
        return indx


    def slitorder_objid_indices(self, slitorder, objid):
        """
        Return the set of indices matching the input slit/order and the input objid
        """
        if self[0].PYPELINE == 'Echelle':
            indx = (self.ECH_ORDERINDX == slitorder) & (self.ECH_OBJID == objid)
        elif self[0].PYPELINE == 'MultiSlit':
            indx = (self.SLITID == slitorder) & (self.OBJID == objid)
        else:
            msgs.error("Should not get here")
        #
        return indx

    def set_names(self):
        for sobj in self.specobjs:
            sobj.set_name()

    def add_sobj(self, sobj):
        """
        Add one or more SpecObj
        The summary table is rebuilt

        Args:
            sobj (SpecObj or list or ndarray):  On or more SpecObj objects

        Returns:


        """
        if isinstance(sobj, specobj.SpecObj):
            self.specobjs = np.append(self.specobjs, [sobj])
        elif isinstance(sobj, (np.ndarray,list)):
            self.specobjs = np.append(self.specobjs, sobj)
        elif isinstance(sobj, SpecObjs):
            for isobj in sobj:
                self.specobjs = np.append(self.specobjs, isobj)

    def remove_sobj(self, index):
        """
        Remove an object

        Args:
            index: int

        Returns:

        """
        msk = np.ones(self.specobjs.size, dtype=bool)
        msk[index] = False
        # Do it
        self.specobjs = self.specobjs[msk]

    def copy(self):
        """
        Generate a copy of self

        Returns:
            SpecObjs

        """
        sobj_copy = SpecObjs(header=self.header)
        for sobj in self.specobjs:
            sobj_copy.add_sobj(sobj.copy())
        return sobj_copy

    def grab_spec_arrays(self, obj_id, DET=None, ECH_ORDER=None, **kwargs):
        # In development
        pass


#    JFH This function is deprecated
#    def set_idx(self):
#        """
#        Set the idx in all the SpecObj
#        Update the summary Table
#
#        Returns:
#
#        """
#        for sobj in self.specobjs:
#            sobj.set_idx()

    def __getitem__(self, item):
        """
        Overload to allow one to pull an attribute or a portion of the
        SpecObjs list

        Args:
            item (:obj:`str`, :obj:`int`, :obj:`slice`)

        Returns:
            The selected items as either an object,
            :class:`pypeit.specobj.SpecObj`, or
            :class:`pypeit.specobjs.SpecObjs`, depending on the input
            item.
        """
        if isinstance(item, str):
            return self.__getattr__(item)
        elif isinstance(item, (int, np.integer)):
            return self.specobjs[item]   # TODO Is this using pointers or creating new data????
        elif (isinstance(item, slice) or  # Stolen from astropy.table
            isinstance(item, np.ndarray) or
            isinstance(item, list) or
            isinstance(item, tuple) and all(isinstance(x, np.ndarray) for x in item)):
            # here for the many ways to give a slice; a tuple of ndarray
            # is produced by np.where, as in t[np.where(t['a'] > 2)]
            # For all, a new table is constructed with slice of all columns
            return SpecObjs(specobjs=self.specobjs[item], header=self.header)

    def __getattr__(self, k):
        """
        Overloaded to generate an array of attribute 'k' from the
        :class:`pypeit.specobj.SpecObj` objects.

        First attempts to grab data from the Summary table, then the list
        """
        if len(self.specobjs) == 0:
            raise ValueError("Empty specobjs")
        try:
            lst = [getattr(specobj, k) for specobj in self.specobjs]
        except ValueError:
            raise ValueError("Attribute does not exist")
        # Recast as an array
        return lst_to_array(lst)

    # Printing
    def __repr__(self):
        txt = '<{:s}:'.format(self.__class__.__name__)
        if self.nobj == 0:
            txt += "Empty SpecObjs"
        else:
            txt += '\n'
            for sobj in self.specobjs:
                txt += '  {} \n'.format(sobj)
        txt += '>'
        return txt

    def __len__(self):
        return len(self.specobjs)

    def build_header(self, head_fitstbl, spectrograph):
        """
        Builds the spec1d file header from the fitstbl meta data and meta data from spectrograph
        Args:
            head_fitstbl (header):
               Header from fitstbl
            spectrograph (object):
               Spectrograph object

        Returns:

        """
        header = fits.PrimaryHDU().header
        try:
            header['MJD-OBS'] = head_fitstbl['mjd']  # recorded as 'mjd' in fitstbl
        except KeyError:
            header['MJD-OBS'] = head_fitstbl['MJD-OBS']
        core_keys = spectrograph.header_cards_for_spec()
        for key in core_keys:
            # Allow for fitstbl vs. header
            try:
                header[key.upper()] = head_fitstbl[key.upper()]
            except KeyError:
                header[key.upper()] = head_fitstbl[key]
        # Specify which pipeline created this file
        header['PYPELINE'] = spectrograph.pypeline
        header['PYP_SPEC'] = (spectrograph.spectrograph, 'PypeIt: Spectrograph name')
        # Observatory
        telescope = spectrograph.telescope
        header['LON-OBS'] = telescope['longitude']
        header['LAT-OBS'] = telescope['latitude']
        header['ALT-OBS'] = telescope['elevation']
        _ = initialize_header(header)
        return header



    def write_to_fits(self, header, outfile, overwrite=True, update_det=None):
        """
        Write the set of SpecObj objects to one multi-extension FITS file

        Args:
            outfile (str):
            header:
            overwrite (bool, optional):
            update_det (int or list, optional):
              If provided, do not clobber the existing file but only update
              the indicated detectors.  Useful for re-running on a subset of detectors

        """
        if os.path.isfile(outfile) and (not overwrite):
            msgs.warn("Outfile exists.  Set overwrite=True to clobber it")
            return

        # If the file exists and update_det is provided, use the existing header
        #   and load up all the other hdus so that we only over-write the ones
        #   we are updating
        if os.path.isfile(outfile) and (update_det is not None):
            hdus, prihdu = save.init_hdus(update_det, outfile)
        else:
            # Build up the Header
            prihdu = fits.PrimaryHDU()
            hdus = [prihdu]
            prihdu.header = header

        ext = len(hdus)-1
        # Loop on the SpecObj objects
        for sobj in self.specobjs:
            if sobj is None:
                continue
            ext += 1
            # Add header keyword
            keywd = 'EXT{:04d}'.format(ext)
            prihdu.header[keywd] = sobj.NAME

            # Table
            shdu = fits.table_to_hdu(sobj._data)
            shdu.name = sobj.NAME
            # Append
            hdus += [shdu]

        # A few more for the header
        prihdu.header['NSPEC'] = len(hdus) - 1
        #prihdu.header['NPIX'] = specObjs.trace_spat.shape[1]
        # Code versions
        #_ = initialize_header(prihdu.header)

        # Finish
        hdulist = fits.HDUList(hdus)
        hdulist.writeto(outfile, overwrite=overwrite)
        msgs.info("Wrote 1D spectra to {:s}".format(outfile))
        return


def lst_to_array(lst, mask=None):
    """
    Simple method to convert a list to an array

    Allows for a list of Quantity objects

    Args:
        lst : list
          Should be number or Quantities
        mask (ndarray of bool, optional):  Limit to a subset of the list.  True=good

    Returns:
        ndarray or Quantity array:  Converted list

    """
    if mask is None:
        mask = np.array([True]*len(lst))
    if isinstance(lst[0], Quantity):
        return Quantity(lst)[mask]
    else:
        return np.array(lst)[mask]



#!/usr/bin/env python
"""
Wrapper to matplotlib to show an arc spectrum
"""
import argparse

def parser(options=None):
    parser = argparse.ArgumentParser(description='Parse')
    parser.add_argument("file", type=str, help="WaveCalib JSON file")
    parser.add_argument("slit", type=str)

    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)
    return args


def main(pargs, unit_test=False):
    """ Shows the spectrum
    """

    import sys
    import pdb

    from matplotlib import pyplot as plt
    from linetools import utils as ltu

    wvcalib = ltu.loadjson(pargs.file)

    # Grab it
    spec = wvcalib[pargs.slit]['spec']

    plt.clf()
    ax = plt.gca()
    ax.plot(spec)
    plt.show()



"""
Module to run tests on BPMImage class
Requires files in Development suite and an Environmental variable
"""
import os

import pytest
import glob
import numpy as np

from pypeit.tests.tstutils import dev_suite_required
from pypeit.spectrographs import util
from pypeit.core import procimg


def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)


@dev_suite_required
def test_keck_lris_red():
    # Spectrograph
    spectrograph = util.load_spectrograph('keck_lris_red')
    #
    example_file = os.path.join(os.getenv('PYPEIT_DEV'), 'RAW_DATA', 'keck_lris_red',
                                'long_600_7500_d560', 'LR.20160216.05529.fits.gz')
    # Get the shape
    det = 2
    # Simple
    bpm = spectrograph.bpm(example_file, det)
    assert np.sum(bpm) > 0


@dev_suite_required
def test_keck_deimos():
    spectrograph = util.load_spectrograph('keck_deimos')
    example_file = os.path.join(os.getenv('PYPEIT_DEV'), 'RAW_DATA', 'keck_deimos', '830G_L_8400',
                                'd0914_0002.fits.gz')
    # Get the shape
    det = 4
    # Simple
    bpm = spectrograph.bpm(example_file, det)
    assert bpm[0,0] == 1




""" Module for LRIS specific codes
"""
import glob
import os
import numpy as np
from astropy.io import fits

from pypeit import msgs
from pypeit.spectrographs import spectrograph
from ..par.pypeitpar import DetectorPar
from .. import telescopes
from pypeit.core import framematch
from pypeit.core import parse
from pypeit.par import pypeitpar
from pkg_resources import resource_filename


from IPython import embed

class GeminiGMOSSpectrograph(spectrograph.Spectrograph):
    """
    Child to handle Gemini/GMOS specific code
    """

    def __init__(self):

        # Get it started
        super(GeminiGMOSSpectrograph, self).__init__()
        self.timeunit = 'isot'  # Synthesizes date+time

    def init_meta(self):
        """
        Generate the meta data dictionary.
        """
        self.meta = {}
        # Required (core)
        self.meta['ra'] = dict(ext=0, card='RA')
        self.meta['dec'] = dict(ext=0, card='DEC')
        self.meta['target'] = dict(ext=0, card='OBJECT')
        self.meta['decker'] = dict(ext=0, card='MASKNAME')
        self.meta['binning'] = dict(card=None, compound=True)
        # TODO: Can we define the card here that compound meta uses to
        # set the binning?  Would be better to have all header cards
        # collected in this function...
#        self.meta['binning'] = dict(ext=1, card='CCDSUM')

        self.meta['mjd'] = dict(ext=0, card='OBSEPOCH')
        self.meta['exptime'] = dict(ext=0, card='EXPTIME')
        self.meta['airmass'] = dict(ext=0, card='AIRMASS')
        # Extras for config and frametyping
        self.meta['dispname'] = dict(ext=0, card='GRATING')
        self.meta['dispangle'] = dict(ext=0, card='CENTWAVE', rtol=1e-5)
        self.meta['dichroic'] = dict(ext=0, card='FILTER1')

        self.meta['datasec'] = dict(ext=1, card='DATASEC')

    def configuration_keys(self):
        """
        Extra keys for defining the configuration

        Returns:

        """
        return ['datasec']


    def compound_meta(self, headarr, meta_key):
        """

        Args:
            headarr: list
            meta_key: str

        Returns:
            value

        """
        if meta_key == 'binning':
            binspatial, binspec = parse.parse_binning(headarr[1]['CCDSUM'])
            binning = parse.binning2string(binspec, binspatial)
            return binning

    def check_frame_type(self, ftype, fitstbl, exprng=None):
        """
        Check for frames of the provided type.
        """
        good_exp = framematch.check_frame_exptime(fitstbl['exptime'], exprng)
        if ftype == 'science':
            return good_exp & (fitstbl['target'] != 'CuAr') & (fitstbl['target'] != 'GCALflat') & (fitstbl['target'] != 'Bias')
            #& (fitstbl['idname'] == 'OBJECT')
        if ftype in ['arc', 'tilt']:
            return good_exp & (fitstbl['target'] == 'CuAr')#& (fitstbl['idname'] == 'ARC')
        if ftype in ['pixelflat', 'trace']:
            return good_exp & (fitstbl['target'] == 'GCALflat')#& (fitstbl['idname'] == 'FLAT')
        if ftype == 'bias':
            return good_exp & (fitstbl['target'] == 'Bias')#& (fitstbl['idname'] == 'BIAS')

        msgs.warn('Cannot determine if frames are of type {0}.'.format(ftype))
        return np.zeros(len(fitstbl), dtype=bool)

    @staticmethod
    def default_pypeit_par():
        """
        Set default parameters for Gemini GMOS reductions.
        """
        par = pypeitpar.PypeItPar()
        par['calibrations']['slitedges']['edge_thresh'] = 20.
        par['calibrations']['slitedges']['fit_order'] = 3

        # 1D wavelength solution
        par['calibrations']['wavelengths']['rms_threshold'] = 0.40  # Might be grating dependent..
        par['calibrations']['wavelengths']['sigdetect'] = 5.  # Doesn't work for reddest chip
        par['calibrations']['wavelengths']['lamps'] = ['CuI', 'ArI', 'ArII']
        par['calibrations']['wavelengths']['method'] = 'full_template'
        par['calibrations']['wavelengths']['nsnippet'] = 1  # 3 detectors splitting is already a lot

        par['calibrations']['tilts']['tracethresh'] = 10.  # Deals with faint CuAr lines

        #   IF YOU CHANGE THIS, YOU WILL NEED TO DEAL WITH THE OVERSCAN GOING ALONG ROWS
        #for key in par['calibrations'].keys():
        #    if 'frame' in key:
        #        par['calibrations'][key]['process']['overscan'] = 'median'

        # Overscan subtract the images
        #par['calibrations']['biasframe']['useframe'] = 'overscan'

        # Alter the method used to combine pixel flats
        par['calibrations']['pixelflatframe']['process']['combine'] = 'median'
        par['calibrations']['pixelflatframe']['process']['sig_lohi'] = [10.,10.]

        # Always correct for flexure
        par['flexure']['method'] = 'boxcar'
        # Splice detectors 1,2,3 when creating sensitivity function
        par['sensfunc']['multi_spec_det'] = [1,2,3]

        # Set the default exposure time ranges for the frame typing
        #par['scienceframe']['exprng'] = [30, None]

        return par

    def config_specific_par(self, scifile, inp_par=None):
        """
        Modify the PypeIt parameters to hard-wired values used for
        specific instrument configurations.

        .. todo::
            Document the changes made!

        Args:
            scifile (str):
                File to use when determining the configuration and how
                to adjust the input parameters.
            inp_par (:class:`pypeit.par.parset.ParSet`, optional):
                Parameter set used for the full run of PypeIt.  If None,
                use :func:`default_pypeit_par`.

        Returns:
            :class:`pypeit.par.parset.ParSet`: The PypeIt paramter set
            adjusted for configuration specific parameter values.
        """
        par = self.__class__.default_pypeit_par() if inp_par is None else inp_par

        headarr = self.get_headarr(scifile)

        # Turn PCA off for long slits
        if 'arcsec' in self.get_meta_value(headarr, 'decker'):
            par['calibrations']['slitedges']['sync_predict'] = 'nearest'

        return par

    def configuration_keys(self):
        """
        Return the metadata keys that defines a unique instrument
        configuration.

        This list is used by :class:`pypeit.metadata.PypeItMetaData` to
        identify the unique configurations among the list of frames read
        for a given reduction.

        Returns:

            list: List of keywords of data pulled from meta
        """
        cfg_keys = super(GeminiGMOSSpectrograph, self).configuration_keys()
        # Add grating tilt
        return cfg_keys+['dispangle']

    def get_rawimage(self, raw_file, det):
        """
        Load up the raw image and generate a few other bits and pieces
        that are key for image processing

        Args:
            raw_file (str):
            det (int):

        Returns:
            tuple:
                raw_img (np.ndarray) -- Raw image for this detector
                hdu (astropy.io.fits.HDUList)
                exptime (float)
                rawdatasec_img (np.ndarray)
                oscansec_img (np.ndarray)

        """
        # Check for file; allow for extra .gz, etc. suffix
        fil = glob.glob(raw_file + '*')
        if len(fil) != 1:
            msgs.error("Found {:d} files matching {:s}".format(len(fil)))

        # Read
        msgs.info("Reading GMOS file: {:s}".format(fil[0]))
        hdu = fits.open(fil[0])
        head0 = hdu[0].header
        head1 = hdu[1].header

        # Number of amplifiers (could pull from DetectorPar but this avoids needing the spectrograph, e.g. view_fits)
        numamp = (len(hdu) - 1) // 3

        # get the x and y binning factors...
        binning = head1['CCDSUM']
        xbin, ybin = [int(ibin) for ibin in binning.split(' ')]

        # First read over the header info to determine the size of the output array...
        datasec = head1['DATASEC']
        x1, x2, y1, y2 = np.array(parse.load_sections(datasec, fmt_iraf=False)).flatten()
        biassec = head1['BIASSEC']
        b1, b2, b3, b4 = np.array(parse.load_sections(biassec, fmt_iraf=False)).flatten()
        nxb = b2 - b1 + 1

        # determine the output array size...
        nx = (x2 - x1 + 1) * numamp + nxb * numamp
        ny = y2 - y1 + 1

        # allocate output array...
        array = np.zeros((nx, ny))
        rawdatasec_img = np.zeros_like(array, dtype=int)
        oscansec_img = np.zeros_like(array, dtype=int)

        if numamp == 2:  # E2V
            if det == 1:  # BLUEST DETECTOR
                order = range(6, 4, -1)
            elif det == 2:  # NEXT
                order = range(3, 5)
            elif det == 3:  # REDDEST DETECTOR
                order = range(1, 3)
        elif numamp == 4:  # Hamamatsu
            if det == 1:  # BLUEST DETECTOR
                order = range(12, 8, -1)
            elif det == 2:  # BLUEST DETECTOR
                order = range(8, 4, -1)
            elif det == 3:  # BLUEST DETECTOR
                order = range(4, 0, -1)
        else:
            embed()

        # insert extensions into master image...
        for kk, jj in enumerate(order):
            # grab complete extension...
            data, overscan, datasec, biassec, x1, x2 = gemini_read_amp(hdu, jj)
            # insert components into output array...
            inx = data.shape[0]
            xs = inx * kk
            xe = xs + inx

            # insert data...
            # Data section
            #section = '[{:d}:{:d},:]'.format(xs * xbin, xe * xbin)  # Eliminate lines
            #dsec.append(section)
            array[xs:xe, :] = np.flipud(data)
            rawdatasec_img[xs:xe, :] = kk+1

            # ; insert postdata...
            xs = nx - numamp * nxb + kk * nxb
            xe = xs + nxb

            #osection = '[{:d}:{:d},:]'.format(xs * xbin, xe * xbin)  # TRANSPOSED FOR WHAT COMES
            #osec.append(osection)
            array[xs:xe, :] = overscan
            oscansec_img[xs:xe, :] = kk+1

        # Need the exposure time
        exptime = hdu[self.meta['exptime']['ext']].header[self.meta['exptime']['card']]
        # Return, transposing array back to orient the overscan properly
        return array.T, hdu, exptime, rawdatasec_img.T, oscansec_img.T


class GeminiGMOSSHamSpectrograph(GeminiGMOSSpectrograph):
    """
    Child to handle Gemini/GMOS-S instrument with Hamamatsu detector
    """
    def __init__(self):

        # Get it started
        super(GeminiGMOSSHamSpectrograph, self).__init__()
        self.spectrograph = 'gemini_gmos_south_ham'
        self.camera = 'GMOS-S'
        self.telescope = telescopes.GeminiSTelescopePar()

        self.detector = [  #  Hamamatsu (since 2014)
            # Detector 1
            DetectorPar(dataext         = 1,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.080,
                        darkcurr        = 0.0,
                        saturation      = 129000.,
                        nonlinear       = 0.95,
                        numamplifiers   = 4,
                        gain            = [1.83]*4,
                        ronoise         = [3.98]*4,
                        suffix          = '_01'
                        ),
            # Detector 2
            DetectorPar(dataext         = 2,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.080,
                        darkcurr        = 0.0,
                        saturation      = 123000.,
                        nonlinear       = 0.95,
                        numamplifiers   = 4,
                        gain            = [1.83]*4,
                        ronoise         = [3.98]*4,
                        suffix          = '_02'
                        ),
            # Detector 3
            DetectorPar(dataext         = 3,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.080,
                        darkcurr        = 0.0,
                        saturation      = 125000.,
                        nonlinear       = 0.95,
                        numamplifiers   = 4,
                        gain            = [1.83]*4,
                        ronoise         = [3.98]*4,
                        suffix          = '_03'
                        ),
        ]
        self.numhead = 13

    @staticmethod
    def default_pypeit_par():
        """
        Set default parameters for XSHOOTER NIR reductions.
        """
        par = GeminiGMOSSpectrograph.default_pypeit_par()
        par['sensfunc']['IR']['telgridfile'] = resource_filename('pypeit', '/data/telluric/TelFit_LasCampanas_3100_26100_R20000.fits')
        return par

    def bpm(self, filename, det, shape=None, msbias=None):
        """ Generate a BPM

        Parameters
        ----------
        det : int, REQUIRED
        **null_kwargs:
           Captured and never used

        Returns
        -------
        badpix : ndarray

        """
        # Get the empty bpm: force is always True
        bpm_img = self.empty_bpm(filename, det, shape=shape)

        # Fill in bad pixels if a master bias frame is provided
        if msbias is not None:
            return self.bpm_frombias(msbias, det, bpm_img)

        if det == 1:
            msgs.info("Using hard-coded BPM for det=1 on GMOSs")

            # TODO: Fix this
            # Get the binning
            hdu = fits.open(filename)
            binning = hdu[1].header['CCDSUM']
            hdu.close()

            # Apply the mask
            xbin = int(binning.split(' ')[0])
            badc = 616//xbin
            bpm_img[badc,:] = 1
        elif det == 2:
            msgs.info("Using hard-coded BPM for det=2 on GMOSs")

            # Get the binning
            hdu = fits.open(filename)
            binning = hdu[1].header['CCDSUM']
            hdu.close()

            # Apply the mask
            xbin = int(binning.split(' ')[0])
            if xbin != 2:
                embed()
            # Up high
            badr = (898*2)//xbin # Transposed
            bpm_img[badr:badr+(8*2)//xbin,:] = 1
            # Down low
            badr = (161*2)//xbin # Transposed
            bpm_img[badr,:] = 1
        elif det == 3:
            msgs.info("Using hard-coded BPM for det=3 on GMOSs")

            # Get the binning
            hdu = fits.open(filename)
            binning = hdu[1].header['CCDSUM']
            hdu.close()

            # Apply the mask
            xbin = int(binning.split(' ')[0])
            if xbin != 2:
                embed()
            badr = (281*2)//xbin # Transposed
            bpm_img[badr:badr+(2*2)//xbin,:] = 1

        return bpm_img

    def config_specific_par(self, scifile, inp_par=None):
        """
        Modify the PypeIt parameters to hard-wired values used for
        specific instrument configurations.

        .. todo::
            Document the changes made!

        Args:
            scifile (str):
                File to use when determining the configuration and how
                to adjust the input parameters.
            inp_par (:class:`pypeit.par.parset.ParSet`, optional):
                Parameter set used for the full run of PypeIt.  If None,
                use :func:`default_pypeit_par`.

        Returns:
            :class:`pypeit.par.parset.ParSet`: The PypeIt paramter set
            adjusted for configuration specific parameter values.
        """
        # Start with instrument wide
        par = super(GeminiGMOSSHamSpectrograph, self).config_specific_par(scifile, inp_par=inp_par)

        if self.get_meta_value(scifile, 'dispname')[0:4] == 'R400':
            par['calibrations']['wavelengths']['reid_arxiv'] = 'gemini_gmos_r400_ham.fits'
        elif self.get_meta_value(scifile, 'dispname')[0:4] == 'B600':
            par['calibrations']['wavelengths']['reid_arxiv'] = 'gemini_gmos_b600_ham.fits'
        #
        return par



class GeminiGMOSNSpectrograph(GeminiGMOSSpectrograph):
    """
    Child to handle Gemini/GMOS-N instrument
    """
    def __init__(self):

        # Get it started
        super(GeminiGMOSNSpectrograph, self).__init__()
        self.telescope = telescopes.GeminiNTelescopePar()
        self.camera = 'GMOS-N'


class GeminiGMOSNHamSpectrograph(GeminiGMOSNSpectrograph):
    """
    Child to handle Gemini/GMOS-N instrument with Hamamatsu detector
    Used since February 2017
    """
    def __init__(self):

        # Get it started
        super(GeminiGMOSNHamSpectrograph, self).__init__()
        self.spectrograph = 'gemini_gmos_north_ham'

        self.detector = [  #  Hamamatsu
            # Detector 1
            DetectorPar(dataext         = 1,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.0807,
                        darkcurr        = 0.0,
                        saturation      = 129000.,
                        nonlinear       = 0.95,
                        numamplifiers   = 4,
                        gain            = [1.63]*4,
                        ronoise         = [4.14]*4,
                        suffix          = '_01'
                        ),
            # Detector 2
            DetectorPar(dataext         = 2,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.0807,
                        darkcurr        = 0.0,
                        saturation      = 123000.,
                        nonlinear       = 0.95,
                        numamplifiers   = 4,
                        gain            = [1.63]*4,
                        ronoise         = [4.14]*4,
                        suffix          = '_02'
                        ),
            # Detector 3
            DetectorPar(dataext         = 3,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.0807,
                        darkcurr        = 0.0,
                        saturation      = 125000.,
                        nonlinear       = 0.95,
                        numamplifiers   = 4,
                        gain            = [1.63]*4,
                        ronoise         = [4.14]*4,
                        suffix          = '_03'
                        ),
        ]
        self.numhead = 13

    def config_specific_par(self, scifile, inp_par=None):
        """
        Modify the PypeIt parameters to hard-wired values used for
        specific instrument configurations.

        .. todo::
            Document the changes made!

        Args:
            scifile (str):
                File to use when determining the configuration and how
                to adjust the input parameters.
            inp_par (:class:`pypeit.par.parset.ParSet`, optional):
                Parameter set used for the full run of PypeIt.  If None,
                use :func:`default_pypeit_par`.

        Returns:
            :class:`pypeit.par.parset.ParSet`: The PypeIt paramter set
            adjusted for configuration specific parameter values.
        """
        # Start with instrument wide
        par = super(GeminiGMOSNHamSpectrograph, self).config_specific_par(scifile, inp_par=inp_par)

        if self.get_meta_value(scifile, 'dispname')[0:4] == 'R400':
            par['calibrations']['wavelengths']['reid_arxiv'] = 'gemini_gmos_r400_ham.fits'
        elif self.get_meta_value(scifile, 'dispname')[0:4] == 'B600':
            par['calibrations']['wavelengths']['reid_arxiv'] = 'gemini_gmos_b600_ham.fits'
        #
        return par

class GeminiGMOSNE2VSpectrograph(GeminiGMOSNSpectrograph):
    """
    Child to handle Gemini/GMOS-N instrument with E2V detector
    Used until February 2017
    """
    def __init__(self):

        # Get it started
        super(GeminiGMOSNE2VSpectrograph, self).__init__()

        self.spectrograph = 'gemini_gmos_north_e2v'

        self.detector = [  #  E2V
            # Detector 1
            DetectorPar(dataext         = 1,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.0728,  # arcsec per pixel
                        darkcurr        = 0.0,
                        saturation      = 110900.,
                        nonlinear       = 0.95,
                        numamplifiers   = 2,
                        gain            = [2.27]*2,
                        ronoise         = [3.32]*2,
                        suffix          = '_01'
                        ),
            # Detector 2
            DetectorPar(dataext         = 2,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.0728,
                        darkcurr        = 0.0,
                        saturation      = 115500.,
                        nonlinear       = 0.95,
                        numamplifiers   = 2,
                        gain            = [2.27]*2,
                        ronoise         = [3.32]*2,
                        suffix          = '_02'
                        ),
            # Detector 3
            DetectorPar(dataext         = 3,  # Not sure this is used
                        specaxis        = 1,
                        xgap            = 0.,
                        ygap            = 0.,
                        ysize           = 1.,
                        platescale      = 0.0728,
                        darkcurr        = 0.0,
                        saturation      = 116700.,
                        nonlinear       = 0.95,
                        numamplifiers   = 2,
                        gain            = [2.27]*2,
                        ronoise         = [3.32]*2,
                        suffix          = '_03'
                        ),
        ]
        self.numhead = 7

    def config_specific_par(self, scifile, inp_par=None):
        """
        Modify the PypeIt parameters to hard-wired values used for
        specific instrument configurations.

        .. todo::
            Document the changes made!

        Args:
            scifile (str):
                File to use when determining the configuration and how
                to adjust the input parameters.
            inp_par (:class:`pypeit.par.parset.ParSet`, optional):
                Parameter set used for the full run of PypeIt.  If None,
                use :func:`default_pypeit_par`.

        Returns:
            :class:`pypeit.par.parset.ParSet`: The PypeIt paramter set
            adjusted for configuration specific parameter values.
        """
        # Start with instrument wide
        par = super(GeminiGMOSNE2VSpectrograph, self).config_specific_par(scifile, inp_par=inp_par)

        if self.get_meta_value(scifile, 'dispname')[0:4] == 'R400':
            par['calibrations']['wavelengths']['reid_arxiv'] = 'gemini_gmos_r400_e2v.fits'
        #
        return par

def gemini_read_amp(inp, ext):
    """
    Read one amplifier of an Gemini GMOS multi-extension FITS image

    Parameters
    ----------
    inp: tuple 
      (str,int) filename, extension
      (hdu,int) FITS hdu, extension

    Returns
    -------
    data
    predata
    postdata
    x1
    y1

    ;------------------------------------------------------------------------
    function lris_read_amp, filename, ext, $
      linebias=linebias, nobias=nobias, $
      predata=predata, postdata=postdata, header=header, $
      x1=x1, x2=x2, y1=y1, y2=y2, GAINDATA=gaindata
    ;------------------------------------------------------------------------
    ; Read one amp from LRIS mHDU image
    ;------------------------------------------------------------------------
    """
    # Parse input
    if isinstance(inp, str):
        hdu = fits.open(inp)
    else:
        hdu = inp

    # get entire extension...
    temp = hdu[ext].data.transpose()
    tsize = temp.shape
    nxt = tsize[0]

    # parse the DETSEC keyword to determine the size of the array.
    header = hdu[ext].header
    detsec = header['DETSEC']
    x1, x2, y1, y2 = np.array(parse.load_sections(detsec, fmt_iraf=False)).flatten()

    # parse the DATASEC keyword to determine the size of the science region (unbinned)
    datasec = header['DATASEC']
    xdata1, xdata2, ydata1, ydata2 = np.array(parse.load_sections(datasec, fmt_iraf=False)).flatten()

    # grab the components...
    data = temp[xdata1-1:xdata2,:]

    # Overscan
    biassec = header['BIASSEC']
    xdata1, xdata2, ydata1, ydata2 = np.array(parse.load_sections(biassec, fmt_iraf=False)).flatten()
    overscan = temp[xdata1-1:xdata2,:]

    # Return
    return data, overscan, datasec, biassec, x1, x2



